"""
badread.py -- Wrapper for BadRead

.. versionadded:: 3.1.5
"""

__all__ = ("BadReadAdapter", "ALL_POSSIBLE_BADREAD_MODELS", "patch_frontend_parser")

import argparse

from labw_utils.typing_importer import List, Final, Mapping, Any
from yasim.llrg_adapter import BaseProcessBasedLLRGAdapter, LLRGInitializationException

ALL_POSSIBLE_BADREAD_MODELS = (
    "nanopore2018",
    "nanopore2020",
    "pacbio2016",
    "verybad",
    "verynice",
)
"""
All possible badread model names

.. versionadded:: 3.1.5
"""


class BadReadAdapter(BaseProcessBasedLLRGAdapter):
    """
    Wrapper for BadRead

    Cmdline Specs::

        if self.model_name == "verybad":
            cmd = [
                llrg_executable_path, "simulate",
                "--reference", self._src_fasta_file_path,
                "--quantity", f"{self._depth}x",
                "--glitches", "1000,100,100",
                "--junk_reads", "5",
                "--random_reads", "5",
                "--chimeras", "10",
                "--identity", "75,90,8",
                "--start_adapter_seq", "",
                "--end_adapter_seq", "",
                *other_args
            ]
        elif self.model_name == "verynice":
            cmd = [
                llrg_executable_path, "simulate",
                "--reference", self._src_fasta_file_path,
                "--quantity", f"{self._depth}x",
                "--error_model", "random",
                "--qscore_model", "ideal",
                "--glitches", "0,0,0",
                "--junk_reads", "0",
                "--random_reads", "0",
                "--chimeras", "0",
                "--identity", "95,100,4",
                "--start_adapter_seq", "",
                "--end_adapter_seq", "",
                *other_args
            ]
        else:
            cmd = [
                llrg_executable_path, "simulate",
                "--reference", self._src_fasta_file_path,
                "--quantity", f"{self._depth}x",
                "--error_model", model_name,
                "--qscore_model", model_name,
                "--start_adapter_seq", "",
                "--end_adapter_seq", "",
                *other_args
            ]

    .. versionadded:: 3.1.5
    """

    @staticmethod
    def validate_params(model_name: str, **kwargs) -> Mapping[str, Any]:
        if model_name not in ALL_POSSIBLE_BADREAD_MODELS:
            raise LLRGInitializationException(f"Illegal model name {model_name}")
        return {}

    llrg_name: Final[str] = "badread"
    _require_integer_depth: Final[bool] = True
    _capture_stdout: Final[bool] = True

    def __init__(
        self,
        *,
        src_fasta_file_path: str,
        dst_fastq_file_prefix: str,
        is_trusted: bool,
        depth: int,
        llrg_executable_path: str,
        model_name: str,
        other_args: List[str],
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param llrg_executable_path: Path to LLRG Executable.
        :param model_name: Name of pre-defined models. See :py:attr:`ALL_POSSIBLE_BADREAD_MODELS`.
        :param is_trusted: Whether to skip input validation test.
        :param other_args: Other arguments to be appended at the bottom of assembled CMD.
        :raise LLRGInitializationException: On error.
        """
        super().__init__(
            src_fasta_file_path=src_fasta_file_path,
            dst_fastq_file_prefix=dst_fastq_file_prefix,
            depth=depth,
            llrg_executable_path=llrg_executable_path,
            is_trusted=is_trusted,
            preserve_intermediate_files=None,  # Not effective.
        )
        if not is_trusted:
            _ = BadReadAdapter.validate_params(model_name=model_name)
        cmd = [
            llrg_executable_path,
            "simulate",
            "--reference",
            self._src_fasta_file_path,
            "--quantity",
            f"{self._depth}x",
        ]
        if model_name == "verybad":
            cmd.extend(
                [
                    "--glitches",
                    "1000,100,100",
                    "--junk_reads",
                    "5",
                    "--random_reads",
                    "5",
                    "--chimeras",
                    "10",
                    "--identity",
                    "75,90,8",
                ]
            )
        elif model_name == "verynice":
            cmd.extend(
                [
                    "--error_model",
                    "random",
                    "--qscore_model",
                    "ideal",
                    "--glitches",
                    "0,0,0",
                    "--junk_reads",
                    "0",
                    "--random_reads",
                    "0",
                    "--chimeras",
                    "0",
                    "--identity",
                    "95,100,4",
                ]
            )
        else:
            cmd.extend(
                [
                    "--error_model",
                    model_name,
                    "--qscore_model",
                    model_name,
                ]
            )
        cmd.extend(["--start_adapter_seq", "", "--end_adapter_seq", "", *other_args])
        self._cmd = cmd

    def _post_execution_hook(self):
        """
        This function is passed since badread pours read into stdout.
        """
        pass

    def _pre_execution_hook(self) -> None:
        """Does not need extra preparation"""
        pass

    @property
    def is_pair_end(self) -> bool:
        return False


def patch_frontend_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """
    Patch argument parser with Badread arguments.

    .. versionadded:: 3.1.5
    """

    parser.add_argument(
        "-m",
        "--model_name",
        required=True,
        help="Badread model name",
        nargs="?",
        type=str,
        action="store",
        choices=ALL_POSSIBLE_BADREAD_MODELS,
    )
    return parser



"""
pbsim3.py -- LLRG adapter for PBSIM v3, a TGS DNA- and RNA-Seq simulator

.. versionadded:: 3.1.5
"""

__all__ = ("main", "create_parser")

import argparse

from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.typing_importer import List
from yasim.helper import llrg
from yasim.helper.rna_seq import bulk_rna_seq_frontend
from yasim.llrg_adapter import pbsim3 as pbsim3


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim pbsim3",
        description=__doc__.splitlines()[1],
    )
    parser = llrg.patch_frontend_parser_public(parser, llrg_name="pbsim3", default_llrg_executable_name="pbsim3")
    parser = llrg.patch_frontend_parser_bulk_rna_seq(parser)
    parser = llrg.patch_frontend_parser_tgs(parser)
    parser = pbsim3.patch_frontend_parser(parser)
    return parser


def main(args: List[str]):
    args, other_args = create_parser().parse_known_args(args)
    bulk_rna_seq_frontend(
        transcriptome_fasta_dir=args.fastas,
        output_fastq_prefix=args.out,
        depth_file_path=args.depth,
        jobs=args.jobs,
        simulator_name="_".join(("pbsim3", args.hmm_model, "ccs" if args.ccs_pass > 1 else "clr"))
        if args.simulator_name is None
        else args.simulator_name,
        adapter_args={
            "hmm_model": args.hmm_model,
            "ccs_num_threads": 1 if args.ccs_pass > 1 else None,
            "samtools_executable_path": args.samtools_path if args.ccs_pass > 1 else None,
            "ccs_executable_path": args.ccs_path if args.ccs_pass > 1 else None,
            "ccs_pass": args.ccs_pass,
            "hmm_method": args.hmm_method,
            "strategy": args.strategy,
            "other_args": other_args,
            "preserve_intermediate_files": args.preserve_intermediate_files,
        },
        assembler_args={
            "truncate_ratio_3p": args.truncate_ratio_3p,
            "truncate_ratio_5p": args.truncate_ratio_5p,
        },
        adapter_class=pbsim3.Pbsim3Adapter,
        is_pair_end=False,
        llrg_executable_path=args.llrg_executable_path,
        not_perform_assemble=args.not_perform_assemble,
    )



"""
merge_pbccs.py -- Merge BAMs created by pbccs.
"""

import argparse
import copy
import glob
import multiprocessing
import os.path
import shutil
import tempfile
import uuid
from multiprocessing.managers import SyncManager
from typing import List

from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.commonutils.stdlib_helper.parallel_helper import (
    ParallelJobExecutor,
    easyexec,
)

__all__ = (
    "create_parser",
    "main",
)

_lh = get_logger(__name__)


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim_scripts merge_pbccs",
        description=__doc__.splitlines()[1],
    )
    parser.add_argument(
        "-o",
        "--out",
        help="Output BAM file",
        type=str,
        required=True,
    )
    parser.add_argument(
        "-e",
        "--pbmerge_path",
        help="Path to pbmerge",
        type=str,
        required=False,
    )
    parser.add_argument(
        "--input_bam_glob",
        type=str,
        required=True,
        help="Glob expression for BAM files that will be merged.",
    )
    return parser


def merge(_d1_path: str, _d2_path: str, _o_path: str, _fns: List[str]) -> None:
    if (
        easyexec(["pbmerge", "-o", _o_path, _d1_path, _d2_path], raise_on_error=False)
        == 0
    ):
        _fns.append(_o_path)
    else:
        _lh.error(f"MERGE %s and %s have errors; discarded", _d1_path, _d2_path)


def main(args: List[str]) -> int:
    argv = create_parser().parse_args(args)
    pbmerge_path = argv.pbmerge_path
    if pbmerge_path is None:
        pbmerge_path = shutil.which("pbmerge")
    if pbmerge_path is None:
        _lh.error("pbmerge not found!")
        return 1
    fns = list(glob.glob(argv.input_bam_glob))
    i = 0
    with tempfile.TemporaryDirectory() as tmpdir:
        while len(fns) > 1:
            sm = SyncManager()
            sm.start()
            i += 1
            print(f"Iter {i} with {len(fns)} files remaining")
            this_fns: List[str] = copy.deepcopy(fns)
            fns_shared = sm.list()
            job_pool = ParallelJobExecutor()
            while len(this_fns) >= 2:
                d1_path, d2_path = this_fns.pop(), this_fns.pop()
                o_path = os.path.join(tmpdir, str(uuid.uuid4()) + ".bam")
                job_pool.append(
                    multiprocessing.Process(
                        target=merge, args=(d1_path, d2_path, o_path, fns_shared)
                    )
                )
            if len(this_fns) == 1:
                fns_shared.append(this_fns.pop())
            job_pool.start()
            job_pool.join()
            fns = list(fns_shared)
            print(f"Remaining: {len(fns)}")
            sm.shutdown()
            sm.join()
        shutil.move(fns[0], argv.out)
    return 0



"""
maf_parser.py -- Parser for LAST aligned/PBSIM MAF.

.. versionadded:: 3.1.5
"""
__all__ = ("MAF_RECORD_REGEX", "MafRecordType", "maf_parse")

import re

from labw_utils.commonutils.lwio.safe_io import get_reader
from labw_utils.commonutils.lwio.tqdm_reader import get_tqdm_line_reader
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import Tuple, Iterable

MafRecordType = Tuple[str, str, str, str]
"""
Name1, Name2, Alignment1, Alignment2

.. versionadded:: 3.1.5
"""

MAF_RECORD_REGEX = re.compile(r"^s +(\S+) +(\d+) +(\d+) +([+-]) +(\d+) +(\S+)$")
"""
Regex to MAF record

.. versionadded:: 3.1.5
"""

_lh = get_logger(__name__)


def maf_parse(maf_path: str, show_tqdm: bool = True) -> Iterable[MafRecordType]:
    """
    Parse MAF into iterable of ``MafRecordType``.

    :param maf_path: Path to input MAF file.
    :param show_tqdm: Whether to show progress bar.

    .. versionadded:: 3.1.5
    """
    num_error = 0
    num_record = 0
    if show_tqdm:
        rf = get_tqdm_line_reader
    else:
        rf = get_reader
    with rf(maf_path) as reader:
        while True:
            line1 = reader.readline()
            if line1 == "":
                break
            if not line1.startswith("s"):
                continue
            else:
                line2 = reader.readline().rstrip("\n")
                line1 = line1.rstrip("\n")
            lm1 = MAF_RECORD_REGEX.match(line1)
            lm2 = MAF_RECORD_REGEX.match(line2)
            if lm1 is None or lm2 is None:
                num_error += 1
                continue
            g1 = lm1.groups()
            g2 = lm2.groups()
            yield g1[0], g2[0], g1[5], g2[5]
            num_record += 1
    _lh.debug("Finished with %d errors and %d records", num_error, num_record)



"""
generate_isoform_depth.py -- Generate Isoform-Level Depth using YASIM V3 API.

.. versionadded:: 3.1.5
"""

__all__ = ("main", "create_parser")

import argparse

import numpy as np

import yasim.helper.depth_io
from labw_utils.bioutils.datastructure.gene_tree import DiploidGeneTree
from labw_utils.bioutils.datastructure.gv.gene import DumbGene
from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.mlutils.ndarray_helper import describe
from labw_utils.typing_importer import List
from yasim.helper import depth
from yasim.helper.frontend import patch_frontend_argument_parser

_lh = get_logger(__name__)


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim generate_isoform_depth",
        description=__doc__.splitlines()[1],
    )
    parser = patch_frontend_argument_parser(parser, "-g")
    parser.add_argument(
        "-o",
        "--out",
        required=True,
        help="Path to output Isoform-Level Depth TSV. Can be compressed.",
        nargs="?",
        type=str,
        action="store",
    )
    parser.add_argument(
        "-d",
        "--depth",
        required=True,
        help="Path to input Gene-Level Depth TSV. Can be compressed.",
        nargs="?",
        type=str,
        action="store",
    )
    parser = patch_frontend_argument_parser(parser, "--low_cutoff")
    parser = patch_frontend_argument_parser(parser, "--high_cutoff_ratio")
    parser.add_argument(
        "--alpha",
        required=False,
        help="Zipf's Coefficient, larger for larger differences",
        nargs="?",
        type=int,
        action="store",
        default=depth.DEFAULT_ALPHA,
    )
    return parser


def main(args: List[str]):
    args = create_parser().parse_args(args)
    gv = DiploidGeneTree.from_gtf_file(args.gtf, gene_implementation=DumbGene)
    gene_level_depth = yasim.helper.depth_io.read_depth(args.depth)
    transcript_level_depth = {}
    for gene in gv.gene_values:
        if gene.gene_id not in gene_level_depth:
            _lh.warning(
                "GEN ISOFORM DEPTH: Gene %s defined in GTF but not gene-level depth",
                gene.gene_id,
            )
        if gene_level_depth[gene.gene_id] == 0:
            for transcript in gene.transcript_values:
                transcript_level_depth[transcript.transcript_id] = 0
            continue
        try:
            this_transcript_level_depth = depth.simulate_isoform_variance_inside_a_gene(
                n=gene.number_of_transcripts,
                mu=gene_level_depth[gene.gene_id],
                low_cutoff=args.low_cutoff,
                alpha=args.alpha,
                high_cutoff_ratio=args.high_cutoff_ratio,
            )
        except depth.GenerationFailureException:
            _lh.error(
                "GEN ISOFORM DEPTH: Generation failed for gene %s -- SKIPPED",
                gene.gene_id,
            )
            continue
        for i, transcript in enumerate(gene.transcript_values):
            transcript_level_depth[transcript.transcript_id] = this_transcript_level_depth[i]
    _lh.info(
        "GEN ISOFORM DEPTH: Generation of isoform-level depth: Final distribution: %s",
        describe(np.array(list(transcript_level_depth.values()))),
    )
    yasim.helper.depth_io.write_depth(transcript_level_depth, args.out, "TRANSCRIPT_ID")



"""
extract_quality_from_maf.py -- Extract per-base alignment status from MAF files.

This script extracts per-base alignment status from MAF files into a TSV file of one line,
which have 4 fields for Insertion, Deletion, Match and Substitution.

Synopsis: python -m yasim_scrips extract_quality_from_maf [MAF1] [[MAF2]...]

Arguments:
    [MAF1] [[MAF2]...] path to MAF files produced by PBSIM3 or LAST aligner.

.. versionadded:: 3.1.5
"""

__all__ = ("main",)

from labw_utils.typing_importer import List

from yasim_scripts.helper.maf_parser import maf_parse


def main(args: List[str]):
    for arg in args:
        all_qual = {"I": 0, "D": 0, "M": 0, "S": 0}
        for maf_record in maf_parse(arg):
            for b1, b2 in zip(maf_record[2].upper(), maf_record[3].upper()):
                if b1 == "-":
                    this_cigar = "I"
                elif b2 == "-":
                    this_cigar = "D"
                elif b1 == "N" or b2 == "N" or b1 == b2:
                    this_cigar = "M"
                else:
                    this_cigar = "S"
                all_qual[this_cigar] += 1
        retl = [arg]
        retl.extend(map(str, all_qual.values()))
        print("\t".join(retl))



"""
featurecounts_to_depth.py -- Convert FeatureCounts Output for NGS and TGS to YASIM input depth.

.. versionadded:: 3.1.5
"""

__all__ = ("main", "create_parser")

import argparse

import pandas as pd

from labw_utils.commonutils.stdlib_helper.argparse_helper import ArgumentParserWithEnhancedFormatHelp
from labw_utils.typing_importer import List
from yasim.helper.depth_io import write_depth


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim_scripts featurecounts_to_depth",
        description=__doc__.splitlines()[1],
    )
    parser.add_argument(
        "-i",
        "--input",
        required=True,
        help="Path to featureCounts/Salmon output TSV",
        nargs="?",
        type=str,
        action="store",
    )
    parser.add_argument(
        "--software",
        required=False,
        help="name of quantification software",
        nargs="?",
        choices=("featureCounts", "Salmon"),
        default="featureCounts",
        type=str,
        action="store",
    )
    parser.add_argument(
        "-o",
        "--out",
        required=True,
        help="Path to output TSV",
        nargs="?",
        type=str,
        action="store",
    )
    parser.add_argument(
        "-f",
        "--feature_name",
        required=False,
        help="Name of output feature",
        choices=("GENE_ID", "TRANSCRIPT_ID"),
        default="TRANSCRIPT_ID",
        nargs="?",
        type=str,
        action="store",
    )
    ngs_tgs_meg = parser.add_mutually_exclusive_group()
    ngs_tgs_meg.add_argument(
        "--read_length",
        required=False,
        help="[For NGS Only] Read length",
        nargs="?",
        type=int,
        default=None,
        action="store",
    )
    ngs_tgs_meg.add_argument(
        "--read_completeness",
        required=False,
        help="[For TGS Only] Mean read completeness",
        nargs="?",
        type=float,
        default=None,
        action="store",
    )
    return parser


def main(args: List[str]) -> int:
    depth = {}
    args = create_parser().parse_args(args)
    fc = pd.read_table(args.input, sep="\t", comment="#")
    if args.read_length is not None:
        if args.software == "featureCounts":

            def get_read_depth(_datum):
                return _datum[-1] * args.read_length / _datum[5]

        else:

            def get_read_depth(_datum):
                return _datum.NumReads * args.read_length / _datum.Length

    elif args.read_completeness is not None:

        def get_read_depth(_datum):
            return _datum[-1] * args.read_completeness

    else:

        def get_read_depth(_datum):
            return _datum[-1]

    for datum in fc.itertuples(index=False):
        real_depth = get_read_depth(datum)
        print(real_depth, datum)
        if real_depth < 0.01:
            continue
        depth[datum[0]] = real_depth
    write_depth(depth, args.out, args.feature_name)
    return 0



"""
TODO docs

.. versionadded:: 3.1.5
.. versionchanged:: 3.1.6
    scRNA-Seq-related functions removed.
"""

import os
import sys
import time
from collections import defaultdict

from labw_utils.commonutils.importer.tqdm_importer import tqdm
from labw_utils.commonutils.lwio import file_system
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.commonutils.stdlib_helper.parallel_helper import ParallelJobExecutor
from labw_utils.typing_importer import Tuple, Mapping, Any, Type, Optional, List
from yasim.helper.depth_io import DepthType, read_depth, DepthParsingException
from yasim.helper.llrg import (
    enhanced_which,
    AssembleDumb,
    AssemblePairEnd,
    AssembleSingleEnd,
    generate_callback,
)
from yasim.llrg_adapter import LLRGInitializationException, BaseLLRGAdapter

_lh = get_logger(__name__)


def validate_adapter_args(
    adapter_args: Mapping[str, Any],
    adapter_class: Type[BaseLLRGAdapter],
    llrg_executable_path: Optional[str] = None,
) -> Mapping[str, Any]:
    """
    TODO docs

    .. versionadded:: 3.1.5
    """
    _lh.info("RNA SEQ: Validating LLRG Adapter parameters...")
    # Validate LLRG Adapter Params
    if llrg_executable_path is not None:
        try:
            llrg_executable_path = enhanced_which(llrg_executable_path)
        except FileNotFoundError:
            _lh.error("RNA SEQ: LLRG not found at %s", llrg_executable_path)
            sys.exit(1)
    adapter_args = dict(adapter_args)
    try:
        adapter_args_update = adapter_class.validate_params(**adapter_args)
    except LLRGInitializationException as e:
        _lh.error("RNA SEQ: Exception %s caught at validating LLRG adapter arguments.", str(e))
        sys.exit(1)
    adapter_args.update(adapter_args_update)
    return adapter_args


def run_rna_seq(
    transcriptome_fasta_dir: str,
    output_fastq_prefix: str,
    depth_data: DepthType,
    jobs: int,
    simulator_name: str,
    adapter_args: Mapping[str, Any],
    assembler_args: Mapping[str, Any],
    adapter_class: Type[BaseLLRGAdapter],
    is_pair_end: bool,
    llrg_executable_path: Optional[str] = None,
    not_perform_assemble: bool = False,
    show_tqdm: bool = True,
):
    """
    TODO docs

    .. versionadded:: 3.1.5
    """
    _lh.info("RNA SEQ: Creating multiprocessing pool and assembler...")
    output_fastq_dir = output_fastq_prefix + ".d"
    os.makedirs(output_fastq_dir, exist_ok=True)
    simulating_pool = ParallelJobExecutor(
        pool_name="Simulating jobs",
        pool_size=jobs,
        delete_after_finish=False,
        show_tqdm=show_tqdm,
    )
    depth_info: List[Tuple[float, str, str]] = []
    """
    Depth information used by LLRG frontend interfaces.
    
    They are: [depth, transcript_id, filename]
    """

    for transcript_id, transcript_depth in depth_data.items():
        filename = os.path.join(transcriptome_fasta_dir, transcript_id + ".fa")
        if not file_system.file_exists(filename):
            _lh.warning("RNA SEQ: FASTA of Isoform %s not exist!", transcript_id)
            continue
        if transcript_depth <= 0:
            _lh.warning("RNA SEQ: Depth of Isoform %s too low!", transcript_id)
            continue
        depth_info.append((transcript_depth, transcript_id, filename))

    # Get assembler arguments
    assembler_full_args = {
        "depth_data": depth_data,
        "output_fastq_prefix": output_fastq_prefix,
        "simulator_name": simulator_name,
        "input_transcriptome_fasta_dir": transcriptome_fasta_dir,
    }
    assembler_full_args.update(assembler_args)

    # Create assembler
    if not_perform_assemble:
        assembler_class = AssembleDumb
    elif is_pair_end:
        assembler_class = AssemblePairEnd
    else:
        assembler_class = AssembleSingleEnd
    assembler = assembler_class(**assembler_full_args)
    assembler.start()

    _lh.info("RNA SEQ: Starting simulation and assembly...")
    for transcript_depth, transcript_id, transcript_filename in tqdm(iterable=depth_info, desc="Submitting jobs..."):
        try:
            sim_thread = adapter_class(
                src_fasta_file_path=transcript_filename,
                dst_fastq_file_prefix=os.path.join(output_fastq_dir, transcript_id),
                depth=transcript_depth,
                llrg_executable_path=llrg_executable_path,  # TODO: Fix this.
                is_trusted=True,
                **adapter_args,
            )
        except LLRGInitializationException as e:
            _lh.warning("Exception %s caught at initialization time", str(e))
            continue
        simulating_pool.append(sim_thread, callback=generate_callback(assembler, transcript_id))

    _lh.info("RNA SEQ: Jobs submitted, waiting...")
    simulating_pool.start()
    simulating_pool.join()

    _lh.info("RNA SEQ: Jobs finished, terminating assembler...")
    assembler.terminate()
    _lh.info("RNA SEQ: Assembler termination signal sent, waiting...")
    while assembler.is_alive():
        _lh.info(
            "RNA SEQ: Assembling %s -- PENDING: %d",
            output_fastq_prefix,
            assembler.n_pending,
        )
        time.sleep(1.0)
    assembler.join()
    _lh.info("RNA SEQ: Assembler finished, retrieving error reports...")

    # Calculating number of exceptions
    exception_dict = defaultdict(lambda: 0)
    for job in simulating_pool.iter_finished_jobs():
        job: BaseLLRGAdapter
        exception_dict[job.exception] += 1
    return exception_dict


def bulk_rna_seq_frontend(
    transcriptome_fasta_dir: str,
    output_fastq_prefix: str,
    depth_file_path: str,
    jobs: int,
    simulator_name: str,
    adapter_args: Mapping[str, Any],
    assembler_args: Mapping[str, Any],
    adapter_class: Type[BaseLLRGAdapter],
    is_pair_end: bool,
    llrg_executable_path: Optional[str] = None,
    not_perform_assemble: bool = False,
) -> int:
    """
    :return: Exit Value

    .. versionadded:: 3.1.5
    """
    adapter_args = validate_adapter_args(
        adapter_args=adapter_args,
        adapter_class=adapter_class,
        llrg_executable_path=llrg_executable_path,
    )
    try:
        depth_data = read_depth(depth_file_path)
    except DepthParsingException:
        _lh.error(f"RNA SEQ: Failed to parse depth file {depth_file_path}")
        return 1

    exception_dict = run_rna_seq(
        transcriptome_fasta_dir=transcriptome_fasta_dir,
        output_fastq_prefix=output_fastq_prefix,
        depth_data=depth_data,
        jobs=jobs,
        simulator_name=simulator_name,
        adapter_args=adapter_args,
        assembler_args=assembler_args,
        adapter_class=adapter_class,
        is_pair_end=is_pair_end,
        llrg_executable_path=llrg_executable_path,
        not_perform_assemble=not_perform_assemble,
        show_tqdm=True,
    )
    _lh.info(f"RNA SEQ: Status of errors: {dict(exception_dict)}")
    _lh.info("RNA SEQ: Simulation finished successfully")
    return 0


def sc_rna_seq_frontend(*args, **kwargs) -> int:
    _ = args, kwargs
    del args, kwargs
    _lh.error("This function was removed.")
    return 1



"""
generate_te_index.py -- NOT FINISHED DO NOT USE Generate Transposon Indices from DFam HDF5 file

.. versionadded:: 3.2.0
"""

__all__ = ("create_parser", "main")

import argparse

from labw_utils.bioutils.datastructure.transposon import TransposonDatabase
from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.typing_importer import List


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim generate_te_index",
        description=__doc__.splitlines()[1],
    )
    parser.add_argument(
        "-i",
        "--src_dfam_h5_file_path",
        required=True,
        help="Input Dfam HDF5 file path. Get one from <https://dfam.org/releases/current/families/>",
        nargs="?",
        type=str,
        action="store",
    )
    parser.add_argument(
        "-o",
        "--dst_te_idx_file_path",
        required=True,
        help="Output TE index path",
        nargs="?",
        type=str,
        action="store",
    )
    parser.add_argument(
        "--dst_consensus_fa_path",
        required=False,
        default=None,
        help="Output TE index path",
        nargs="?",
        type=str,
        action="store",
    )
    parser.add_argument(
        "--dst_hmm_path",
        required=False,
        default=None,
        help="Output TE index path",
        nargs="?",
        type=str,
        action="store",
    )
    parser.add_argument(
        "-t",
        "--txid",
        required=False,
        help="NCBI taxon ID to subset.",
        nargs="?",
        type=int,
        action="store",
        default=9606,
    )
    return parser


def main(args: List[str]):
    args = create_parser().parse_args(args)
    TransposonDatabase.convert_dfam_hdf5(
        src_dfam_hdf5_file_path=args.src_dfam_h5_file_path,
        dst_index_file_path=args.dst_te_idx_file_path,
        dst_consensus_fa_path=args.dst_consensus_fa_path,
        dst_hmm_path=args.dst_hmm_path,
        with_tqdm=True,
        required_txid=args.txid,
    )



"""
generate_gene_depth.py -- Generate Gene-Level Depth using YASIM V3 API.

.. versionadded:: 3.1.5
"""

__all__ = ("main", "create_parser")

import argparse

from labw_utils.bioutils.datastructure.gene_tree import DiploidGeneTree
from labw_utils.bioutils.datastructure.gv.gene import DumbGene
from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List
from yasim.helper import depth, depth_io
from yasim.helper.frontend import patch_frontend_argument_parser

_lh = get_logger(__name__)


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim generate_gene_depth",
        description=__doc__.splitlines()[1],
    )
    parser = patch_frontend_argument_parser(parser, "-g")
    parser.add_argument(
        "-o",
        "--out",
        required=True,
        help="Path to output Gene-Level depth TSV. Can be compressed.",
        nargs="?",
        type=str,
        action="store",
    )
    parser.add_argument(
        "-d",
        "--mu",
        required=False,
        help="Average depth.",
        nargs="?",
        type=float,
        action="store",
        default=depth.DEFAULT_MU,
    )
    parser = patch_frontend_argument_parser(parser, "--low_cutoff")
    parser = patch_frontend_argument_parser(parser, "--high_cutoff_ratio")
    return parser


def main(args: List[str]) -> int:
    args = create_parser().parse_args(args)
    gv = DiploidGeneTree.from_gtf_file(args.gtf, gene_implementation=DumbGene)
    try:
        dge_data = depth.simulate_gene_level_depth_gmm(
            gene_names=gv.gene_ids,
            mu=args.mu,
            low_cutoff=args.low_cutoff,
            high_cutoff_ratio=args.high_cutoff_ratio,
        )
    except depth.GenerationFailureException:
        _lh.error("GEN GENE DEPTH: Generation failed! You may try again or use larger `mu`.")
        return 1
    depth_io.write_depth(dge_data, args.out, feature_name="GENE_ID")



"""
pbsim.py -- Wrapper of PBSIM.

.. versionadded:: 3.1.5
"""

__all__ = ("PbsimAdapter", "PBSIM_DIST_DIR_PATH", "patch_frontend_parser")

import argparse
import glob
import os

from labw_utils.typing_importer import List, Final, Mapping, Any
from yasim.helper.frontend import patch_frontend_argument_parser
from yasim.llrg_adapter import BaseProcessBasedLLRGAdapter, automerge

PBSIM_DIST_DIR_PATH = os.path.join(os.path.dirname(__file__), "pbsim_dist")
"""
Where pbsim stores its models

.. versionadded:: 3.1.5
"""


class PbsimAdapter(BaseProcessBasedLLRGAdapter):
    """
    Wrapper of PBSIM.

    Cmdline Spec::

        if self.is_ccs:
            cmd = [
                llrg_executable_path,
                "--prefix", self._tmp_dir,
                "--_depth", str(self._depth),
                "--data-type", "CCS",
                "--model_qc", os.path.join(PBSIM_DIST_DIR_PATH, "model_qc_ccs"),
                *other_args,
                self._src_fasta_file_path
            ]
        else:
            cmd = [
                llrg_executable_path,
                "--prefix", self._tmp_dir,
                "--_depth", str(self._depth),
                "--data-type", "CLR",
                "--model_qc", os.path.join(PBSIM_DIST_DIR_PATH, "model_qc_clr"),
                *other_args,
                self._src_fasta_file_path
            ]

    .. versionadded:: 3.1.5
    """

    llrg_name: Final[str] = "pbsim"
    _require_integer_depth: Final[bool] = False
    _capture_stdout: Final[bool] = False

    @staticmethod
    def validate_params(**kwargs) -> Mapping[str, Any]:
        return {}

    def __init__(
        self,
        *,
        src_fasta_file_path: str,
        dst_fastq_file_prefix: str,
        depth: int,
        llrg_executable_path: str,
        is_trusted: bool,
        is_ccs: bool,
        preserve_intermediate_files: bool,
        other_args: List[str],
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param llrg_executable_path: Path to LLRG Executable.
        :param other_args: Other arguments to be appended at the bottom of assembled CMD.
        :param is_trusted: Whether to skip input validation test.
        :param is_ccs: Whether to simulate CCS (True) or CLR (False) data.
        :raise LLRGInitializationException: On error.
        """
        super().__init__(
            src_fasta_file_path=src_fasta_file_path,
            dst_fastq_file_prefix=dst_fastq_file_prefix,
            depth=depth,
            llrg_executable_path=llrg_executable_path,
            is_trusted=is_trusted,
            preserve_intermediate_files=preserve_intermediate_files,
        )

        cmd = [
            llrg_executable_path,
            "--prefix",
            os.path.join(self._tmp_dir, "tmp"),
            "--depth",
            str(self._depth),
        ]
        if is_ccs:
            cmd.extend(
                [
                    "--data-type",
                    "CCS",
                    "--model_qc",
                    os.path.join(PBSIM_DIST_DIR_PATH, "model_qc_ccs"),
                ]
            )
        else:
            cmd.extend(
                [
                    "--data-type",
                    "CLR",
                    "--model_qc",
                    os.path.join(PBSIM_DIST_DIR_PATH, "model_qc_clr"),
                ]
            )
        cmd.extend([*other_args, self._src_fasta_file_path])
        self._cmd = cmd

    def _pre_execution_hook(self) -> None:
        """Does not need extra preparation"""
        pass

    def _post_execution_hook(self):
        automerge(
            glob.glob(os.path.join(self._tmp_dir, "tmp_????.fastq")),
            self._dst_fastq_file_prefix + ".fq",
        )

    @property
    def is_pair_end(self) -> bool:
        return False


def patch_frontend_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """
    Patch argument parser with pbsim arguments.

    .. versionadded:: 3.1.5
    """
    parser.add_argument(
        "-c",
        "--ccs",
        required=False,
        help="Simulate CCS instead of CLR",
        action="store_true",
    )
    parser = patch_frontend_argument_parser(parser, "--preserve_intermediate_files")
    return parser



"""
yasim -- Yet Another SIMulator for Alternative Splicing and Realistic Gene Expression Profile

.. versionadded:: 3.1.5
"""

__version__ = "3.2.0"

description = __doc__.splitlines()[1]

try:
    import labw_utils

    _labw_utils_version = labw_utils.__version__

except ImportError as e:
    raise e

from labw_utils import UnmetDependenciesError

try:
    import numpy
    import numpy.typing as npt
except ImportError as e:
    raise UnmetDependenciesError("numpy") from e
try:
    import pandas
except ImportError as e:
    raise UnmetDependenciesError("pandas") from e

try:
    import joblib
except ImportError as e:
    raise UnmetDependenciesError("joblib") from e

try:
    import scipy
except ImportError as e:
    raise UnmetDependenciesError("scipy") from e



"""
Fit Gaussian Mixture by YUAN Ruihong

.. versionadded:: 3.1.5
"""

import argparse
import re
from random import random

import numpy as np
import pandas as pd
from scipy.stats import norm

from labw_utils import UnmetDependenciesError
from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.mlutils.ndarray_helper import describe
from labw_utils.typing_importer import List
from yasim.helper.gmm import GaussianMixture1D
from yasim.helper.plot_utils import plot

try:
    import pyarrow as _
except ImportError as e:
    raise UnmetDependenciesError("pyarrow") from e

_lh = get_logger()


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim_scripts fit_gmm",
        description=__doc__.splitlines()[1],
    )
    parser.add_argument(
        "data",
        help="Dataset in Apache Parquet Format",
    )
    parser.add_argument(
        "-c",
        "--num_components",
        type=int,
        required=False,
        default=2,
        help="Number of components." "Larger value can capture more details but be at a rick of over fit.",
    )
    parser.add_argument(
        "--col_regex",
        type=str,
        required=False,
        default="SRR.*",
        help="Regular expression for column names that will be modelled.",
    )
    parser.add_argument(
        "--filter_zero",
        action="store_true",
        help="Filter zero.",
    )
    parser.add_argument(
        "--scale",
        action="store_true",
        help="Scale fata to [1, 5001].",
    )
    parser.add_argument(
        "--num_iters",
        type=int,
        required=False,
        default=100,
        help="Number of EM Iterations. Most samples will converge before 100.",
    )
    return parser


def main(args: List[str]):
    args = create_parser().parse_args(args)
    _lh.info("GMM: Processing data...")
    if args.data == "dummy":
        mu1, sigma1 = random() * 100, random() * 10
        mu2, sigma2 = random() * 100, random() * 10
        p = random()
        model1 = norm(mu1, sigma1)
        model2 = norm(mu2, sigma2)
        data = np.append(model1.rvs(size=int(p * 200)), model2.rvs(size=int((1 - p) * 200)))
        print(f"Ground truth is {p}*N({mu1}, {sigma1}) + {1 - p}*N({mu2}, {sigma2})")
    else:
        df = pd.read_parquet(args.data)
        col_names = df.columns[list(map(lambda x: re.match(args.col_regex, x) is not None, df.columns))]
        df: pd.DataFrame = df[col_names]
        df = df.fillna(0)
        if args.scale:
            df = df.apply(lambda x: np.log10(x / max(x.max(), 1e-9) * 5000 + 1), axis="columns")

        data = np.asarray(df.to_numpy()).ravel()
        if args.filter_zero:
            data = data[data != 0]
    print(describe(data))
    # plot(data)

    _lh.info("GMM: Fitting START")
    # auto_select_num_components(data, 2, 8, args.num_iters)
    model = GaussianMixture1D(n_components=args.num_components, n_iter=args.num_iters).fit(data)
    _lh.info("GMM: Fitting FIN")

    plot(data, model)
    print(model.export())



"""
generate_depth_v2.py -- Generate Isoform-Level Depth using YASIM V2 API.

.. versionadded:: 0.2.7
    Named as generate_depth

.. versionchanged:: 3.0.0
    Renamed to generate_depth_v2

.. versiondeprecated:: 3.1.5
    Use V3API instead.
"""

__all__ = ("main", "create_parser")

import argparse

from labw_utils.bioutils.datastructure.gene_tree import DiploidGeneTree
from labw_utils.bioutils.datastructure.gv.gene import DumbGene
from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List
from yasim.helper import depth, depth_io
from yasim.helper.frontend import patch_frontend_argument_parser

_lh = get_logger(__name__)


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim generate_depth_v2",
        description=__doc__.splitlines()[1],
    )
    parser = patch_frontend_argument_parser(parser, "-g")
    parser.add_argument(
        "-o",
        "--out",
        required=True,
        help="Path to output Isoform-Level depth TSV. Can be compressed.",
        nargs="?",
        type=str,
        action="store",
    )
    parser.add_argument(
        "-d",
        "--mu",
        required=False,
        help="Average depth",
        nargs="?",
        type=float,
        action="store",
        default=100,
    )
    return parser


def main(args: List[str]):
    _lh.warning(
        "DEPRECATION WARNING: "
        "The V2 API had been deprecated. Use `generate_gene_depth` and `generate_isoform_depth` instead"
    )
    args = create_parser().parse_args(args)
    gv = DiploidGeneTree.from_gtf_file(args.gtf, gene_implementation=DumbGene)
    dge_data = depth.simulate_depth_gmm_v2(isoform_names=gv.transcript_ids, mu=args.mu)
    depth_io.write_depth(dge_data, args.out, feature_name="TRANSCRIPT_ID")



"""
generate_gene_te_fusion.py -- NOT FINISHED DO NOT USE

.. versionadded:: 3.2.0
"""

__all__ = ("main", "create_parser")

import argparse
import random

from labw_utils.bioutils.datastructure.fasta_view import FastaViewFactory
from labw_utils.bioutils.datastructure.gene_tree import DiploidGeneTree
from labw_utils.bioutils.datastructure.gv.gene import DumbGene
from labw_utils.bioutils.datastructure.transposon import TransposonDatabase
from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.typing_importer import List
from yasim.helper.depth import DEFAULT_MU
from yasim.helper.frontend import patch_frontend_argument_parser
from yasim.helper.translation_instruction import TranslationInstruction

rdg = random.SystemRandom()


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim insert_transposon",
        description=__doc__.splitlines()[1],
    )
    parser = patch_frontend_argument_parser(parser, "-g")
    parser = patch_frontend_argument_parser(parser, "-f")
    parser.add_argument(
        "--tedb",
        type=str,
        help="Path to TE database index.",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-o",
        "--out",
        type=str,
        help="Path to base path where result file should be written to.",
        required=True,
    )
    parser.add_argument(
        "-n",
        "--nseqs",
        required=True,
        help="Number of contigs to generate",
        nargs="?",
        type=int,
        action="store",
    )
    parser.add_argument(
        "-d",
        "--mu",
        required=False,
        help="Average depth",
        nargs="?",
        type=float,
        action="store",
        default=DEFAULT_MU,
    )
    parser = patch_frontend_argument_parser(parser, "--low_cutoff")
    parser = patch_frontend_argument_parser(parser, "--high_cutoff_ratio")
    return parser


def main(args: List[str]):
    argv = create_parser().parse_args(args)
    ti = TranslationInstruction.generate(
        n=argv.nseqs,
        mu=argv.mu,
        fav=FastaViewFactory(argv.fasta),
        gt=DiploidGeneTree.from_gtf_file(argv.gtf, gene_implementation=DumbGene),
        tedb=TransposonDatabase.load(argv.tedb) if argv.tedb is not None else None,
        low_cutoff=argv.low_cutoff,
        high_cutoff_ratio=argv.high_cutoff_ratio,
    )
    ti.to_fasta(argv.out + ".fa")
    ti.to_depth(argv.out + ".depth.tsv")
    ti.to_json(argv.out + ".json")



"""
art.py -- LLRG adapter for ART, a NGS DNA-Seq simulator

.. versionadded:: 3.1.5
"""

__all__ = ("main", "create_parser")

import argparse

from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.typing_importer import List
from yasim.helper import llrg
from yasim.helper.rna_seq import bulk_rna_seq_frontend
from yasim.llrg_adapter import art


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim art",
        description=__doc__.splitlines()[1],
    )
    parser = llrg.patch_frontend_parser_public(
        parser,
        llrg_name="art",
        default_llrg_executable_name="art_illumina",
    )
    parser = llrg.patch_frontend_parser_bulk_rna_seq(parser)
    parser = art.patch_frontend_parser(parser)
    return parser


def main(args: List[str]) -> int:
    args, other_args = create_parser().parse_known_args(args)

    return bulk_rna_seq_frontend(
        transcriptome_fasta_dir=args.fastas,
        output_fastq_prefix=args.out,
        depth_file_path=args.depth,
        jobs=args.jobs,
        simulator_name="art_illumina" if args.simulator_name is None else args.simulator_name,
        adapter_args={
            "other_args": other_args,
            "sequencer_name": args.sequencer_name,
            "read_length": args.read_length,
            "pair_end_fragment_length_mean": args.pair_end_fragment_length_mean,
            "pair_end_fragment_length_std": args.pair_end_fragment_length_std,
            "is_pair_end": args.is_pair_end,
            "preserve_intermediate_files": args.preserve_intermediate_files,
        },
        assembler_args={},
        adapter_class=art.ArtAdapter,
        is_pair_end=args.is_pair_end,
        llrg_executable_path=args.llrg_executable_path,
        not_perform_assemble=args.not_perform_assemble,
    )



"""
gmm.py -- Gaussian Mixture Model bu YUAN Ruihong

.. versionadded:: 3.1.5
"""

__all__ = ("GaussianMixture1D",)

import itertools
import multiprocessing
import secrets

import numpy as np
import numpy.typing as npt
import tqdm
from scipy.integrate import quad
from scipy.stats import norm

from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.commonutils.stdlib_helper.parallel_helper import parallel_map
from labw_utils.typing_importer import (
    Optional,
    Union,
    Iterable,
    Tuple,
    List,
    Sequence,
    SequenceProxy,
    Any,
)

_lh = get_logger(__name__)


class GaussianMixture1D:
    """
    TODO docs

    .. versionadded:: 3.1.5
    """

    _n_components: int
    _n_iter: int
    _precision: float
    _weights: npt.NDArray
    _mu: npt.NDArray
    _sigma: npt.NDArray

    def __init__(
        self,
        n_components: int = 1,
        n_iter: int = 100,
        init_weights: Optional[npt.ArrayLike] = None,
        init_mus: Optional[npt.ArrayLike] = None,
        init_sigmas: Optional[npt.ArrayLike] = None,
        precision: float = 1e-4,
    ) -> None:
        self._n_components = n_components
        self._n_iter = n_iter
        self._precision = precision
        self._weights = init_weights or np.repeat(1 / n_components, n_components)
        self._mu = init_mus or np.zeros(n_components)
        self._sigma = init_sigmas or np.ones(n_components)

    def fit(self, data: npt.ArrayLike, show_progress_bar: bool = True):
        n = len(data)
        rdg = secrets.SystemRandom()

        # Use k-means to cluster points
        centers = np.array(rdg.choices(data, k=self._n_components))
        categories = np.repeat(0, n)
        bags: List[List[Union[int, float]]] = [[]] * self._n_components
        knn_range = range(self._n_iter)
        if show_progress_bar:
            knn_range = tqdm.tqdm(iterable=knn_range, desc="KNN")
        for _ in knn_range:
            converged = True
            for j in range(self._n_components):
                bags[j] = []

            for i in range(n):
                x = data[i]
                category = categories[i]
                least_dist = abs(x - centers[category])
                for j in range(self._n_components):
                    center = centers[j]
                    dist = abs(x - center)
                    if dist < least_dist:
                        converged = False
                        category = j  # assign to the nearest center
                        least_dist = dist
                categories[i] = category
                bags[category].append(x)

            for j in range(self._n_components):
                bag = bags[j]
                if len(bag) != 0:
                    centers[j] = np.mean(bag)

            if converged:
                break

        self._mu = centers
        for j in range(self._n_components):
            bag = bags[j]
            if len(bag) != 0:
                self._sigma[j] = np.std(bag)
                self._weights[j] = len(bag)
            else:
                self._sigma[j] = 0
                self._weights[j] = 0
        self._weights /= n

        last_ll = np.NINF
        densities = np.zeros((self._n_components, n))
        em_range = range(self._n_iter)
        if show_progress_bar:
            em_range = tqdm.tqdm(iterable=em_range, desc="EM")
        _last_range = self.export()
        for _ in em_range:
            # E step: compute ownership weights
            for j in range(self._n_components):
                model = norm(loc=self._mu[j], scale=self._sigma[j])
                densities[j, :] = model.pdf(data)

            a = np.transpose(np.multiply(np.transpose(densities), self._weights))
            b = np.transpose(self._weights) @ densities
            gamma = a / b

            # M step: compute means, variances and weights
            denom = np.sum(gamma, axis=1)  # row sum
            self._mu = (gamma @ data) / denom

            for j in range(self._n_components):
                for i in range(n):
                    gamma[j, i] *= (data[i] - self._mu[j]) ** 2

            self._sigma = np.sqrt(np.sum(gamma, axis=1) / denom)
            self._weights = denom / n

            ll = self.log_likelihood(data)
            if 0 <= ll - last_ll < self._precision:
                break
            last_ll = ll
            current_export = np.array(list(itertools.chain(*self.export())), dtype=float)
            if np.sum(np.bitwise_or(np.isnan(current_export), np.isinf(current_export))) != 0:
                _lh.warning("GMM: NaN observed, will rerun last successful result")
                return self.import_model(_last_range)
            else:
                _last_range = list(self.export())
        return self

    def log_likelihood(self, data):
        return np.sum(self.logpdf(data))

    def pdf(self, x: Union[int, float, npt.ArrayLike]) -> float:
        if isinstance(x, (int, float)):
            result = 0.0
        else:
            result = np.zeros(len(x))

        for j in range(self._n_components):
            model = norm(loc=self._mu[j], scale=self._sigma[j])
            result += self._weights[j] * model.pdf(x)
        return result

    def logpdf(self, x: Union[int, float, npt.ArrayLike]) -> float:
        if isinstance(x, (int, float)):
            result = 0.0
        else:
            result = np.zeros(len(x))

        for j in range(self._n_components):
            model = norm(loc=self._mu[j], scale=self._sigma[j])
            result += self._weights[j] * model.logpdf(x)
        return result

    def lintrans(self, a: float, b: float = 0):
        self._mu *= a
        self._mu += b
        self._sigma *= a
        return self

    def positive_mean(self, start: float = 0):
        models = []
        for j in range(self._n_components):
            models.append(norm(self._mu[j], self._sigma[j]))

        def f(x):
            result = 0.0
            for _j in range(self._n_components):
                result += self._weights[_j] * models[_j].pdf(x)
            return result * x

        return quad(f, start, np.inf)[0]

    def mean(self) -> float:
        return np.average(self._mu, weights=self._weights)

    def rvs(self, size: int = 1) -> npt.ArrayLike:
        indices = list(range(self._n_components))

        def _rvs(_: Any):
            rdg = secrets.SystemRandom()
            k = rdg.choices(indices, weights=self._weights)
            return rdg.normalvariate(mu=self._mu[k], sigma=self._sigma[k])

        result = np.array(
            list(parallel_map(_rvs, range(size), n_jobs=multiprocessing.cpu_count())),
            dtype=float,
        )
        return result

    def export(self) -> Sequence[Tuple[float, float, float]]:
        retl = []
        for j in range(self._n_components):
            retl.append((self._weights[j], self._mu[j], self._sigma[j]))
        return SequenceProxy(retl)

    @classmethod
    def import_model(cls, model: Iterable[Tuple[float, float, float]]):
        model = list(model)
        new_instance = cls(n_components=len(model))
        for j in range(len(model)):
            (
                new_instance._weights[j],
                new_instance._mu[j],
                new_instance._sigma[j],
            ) = model[j]
        return new_instance

    def __repr__(self):
        model_desc = "Model ="
        for j in self.export():
            model_desc += f" + {j[0]}*N({j[1]}, {j[2]})"
        return model_desc

    def __str__(self):
        return repr(self)



"""
dtgs.py -- Wrapper of dTGS simulator (Dumb Third-Generation Sequencing Simulator).

.. versionadded:: 3.1.5
"""

__all__ = ("DTGSAdapter",)

from labw_utils.bioutils.datastructure.fasta_view import FastaViewFactory
from labw_utils.bioutils.parser.fastq import FastqWriter
from labw_utils.bioutils.record.fastq import FastqRecord
from labw_utils.typing_importer import List, Union, Final, Mapping, Any
from yasim.llrg_adapter import BaseFunctionBasedLLRGAdapter


class DTGSAdapter(BaseFunctionBasedLLRGAdapter):
    """
    Wrapper of dTGS simulator (Dumb Third-Generation Sequencing Simulator).

    .. versionadded:: 3.1.5
    """

    def llrg_func(self):
        fav = FastaViewFactory(self.input_fasta, show_tqdm=False)
        with FastqWriter(self.output_fastq_prefix + ".fq") as fastq_writer:
            i = 0
            for chr_name in fav.chr_names:
                chr_seq = fav.sequence(chr_name)
                for _ in range(self.depth):
                    seqlen = len(chr_seq)
                    fastq_writer.write(FastqRecord(seq_id=str(i), sequence=chr_seq, quality="K" * seqlen))
                    i += 1

    llrg_name: Final[str] = "dtgs"
    _require_integer_depth: Final[bool] = True

    @staticmethod
    def validate_params(**kwargs) -> Mapping[str, Any]:
        return {}

    def __init__(
        self,
        *,
        src_fasta_file_path: str,
        dst_fastq_file_prefix: str,
        depth: Union[int, float],
        is_trusted: bool,
        other_args: List[str],
        **kwargs,
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param other_args: Other arguments to be appended at the bottom of assembled CMD.
        :param is_trusted: Whether to skip input validation test.
        :raise LLRGInitializationException: On error.
        """
        _ = other_args, kwargs
        super().__init__(
            src_fasta_file_path=src_fasta_file_path,
            dst_fastq_file_prefix=dst_fastq_file_prefix,
            depth=depth,
            is_trusted=is_trusted,
            preserve_intermediate_files=None,  # Not effective.
        )

    def _pre_execution_hook(self) -> None:
        """Does not need extra preparation"""
        pass

    def _post_execution_hook(self):
        """Does not need extra preparation"""
        pass

    @property
    def is_pair_end(self) -> bool:
        return False



"""
depth_io.py -- GEP Datastructure and Utils

Here contains Datastructure and I/O utilities for Gene Expression Profile (GEP).

.. versionadded:: 3.1.5
"""

__all__ = ("DepthType", "write_depth", "read_depth", "DepthParsingException")

from labw_utils.commonutils.importer.tqdm_importer import tqdm
from labw_utils.commonutils.lwio.safe_io import get_writer, get_reader
from labw_utils.commonutils.lwio.tqdm_reader import get_tqdm_line_reader
from labw_utils.typing_importer import Dict, Literal

DepthType = Dict[str, float]
"""
DGE type, is transcript_id -> coverage

.. versionadded:: 3.1.5
"""


class DepthParsingException(RuntimeError):
    ...


def write_depth(
    depth_data: DepthType,
    dst_depth_file_path: str,
    feature_name: Literal["GENE_ID", "TRANSCRIPT_ID"],
    show_tqdm: bool = True,
):
    """
    Write Depth information to file

    :param depth_data: Data in depth.
    :param dst_depth_file_path: Output TSV path. Can be compressed.
    :param feature_name: Name of deature. Should be ``GENE_ID`` or ``TRANSCRIPT_ID``.
    :param show_tqdm: Whether to show a progress bar.

    .. versionadded:: 3.1.5
    """
    with get_writer(dst_depth_file_path) as writer:
        writer.write(f"{feature_name}\tDEPTH\n")
        it = depth_data.items()
        if show_tqdm:
            it = tqdm(it, desc=f"Writing to {dst_depth_file_path}")
        for transcript_id, d in it:
            writer.write(f"{transcript_id}\t{d}\n")


def read_depth(src_depth_file_path: str, show_tqdm: bool = True) -> DepthType:
    """
    Read depth information from a file.

    :param src_depth_file_path: Path to source depth file. Can be compressed.
    :param show_tqdm: Whether to show a progress bar.
    :return: Retrived depth information

    .. versionadded:: 3.1.5
    """
    try:
        retd = {}
        if show_tqdm:
            reader = get_tqdm_line_reader(src_depth_file_path)
        else:
            reader = get_reader(src_depth_file_path)
        _ = next(reader)  # Skip line 1
        for line in reader:
            line = line.strip()
            lkv = line.split("\t")
            retd[lkv[0]] = float(lkv[1])
        reader.close()
        return retd
    except (OSError, KeyError, IndexError) as e:
        raise DepthParsingException(f"Cannot parse {src_depth_file_path}") from e



"""
TODO docs

.. versionadded:: 3.1.6
"""


from labw_utils.commonutils.libfrontend import setup_frontend

from yasim import __version__ as yasim_ver, description

if __name__ == "__main__":
    setup_frontend(
        f"{__package__}._main",
        description,
        yasim_ver,
    )



"""
generate_as_events.py -- Generate Alternative Splicing Events from Reference genome using YASIM V3 API.

.. versionadded:: 3.1.5
"""

__all__ = ("main", "create_parser")

import argparse

from labw_utils.bioutils.datastructure.gene_tree import DiploidGeneTree, GeneTree
from labw_utils.bioutils.datastructure.gv.gene import DumbGene
from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List
from yasim.helper.as_events import ASManipulator
from yasim.helper.frontend import patch_frontend_argument_parser

_lh = get_logger(__name__)


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim generate_as_events",
        description=__doc__.splitlines()[1],
    )
    parser = patch_frontend_argument_parser(parser, "-g")
    parser = patch_frontend_argument_parser(parser, "-f")
    parser.add_argument(
        "-c",
        "--complexity",
        required=True,
        help="Transcriptome Complexity Index, should be an integer between 1 and 9.",
        nargs="?",
        type=int,
        action="store",
    )
    parser.add_argument(
        "-o",
        "--out",
        required=True,
        help="Path to output genome annotation with Ground-Truth AS Events in GTF.",
        nargs="?",
        type=str,
        action="store",
    )
    return parser


def main(args: List[str]):
    args = create_parser().parse_args(args)
    # TODO: Change this to DiploidGeneTree. Now it failed with NotImplementedError.
    gv = GeneTree.from_gtf_file(args.gtf, gene_implementation=DumbGene)
    asm = ASManipulator(gv=gv)
    asm.run("ce", args.complexity)
    asm.to_file(args.out)



"""
art.py -- ART adapter for Illumina sequencing.

.. versionadded:: 3.1.5
"""

__all__ = ("AVAILABLE_ILLUMINA_ART_SEQUENCER", "ArtAdapter", "patch_frontend_parser")

import argparse
import os

from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List, Tuple, Union, Final, Any, Mapping
from yasim.helper.frontend import patch_frontend_argument_parser
from yasim.llrg_adapter import (
    BaseProcessBasedLLRGAdapter,
    autocopy,
    LLRGInitializationException,
)

AVAILABLE_ILLUMINA_ART_SEQUENCER: Mapping[str, Tuple[str, List[int]]] = {
    "GA1": ("GenomeAnalyzer I", [36, 44]),
    "GA2": ("GenomeAnalyzer II", [50, 75]),
    "HS10": ("HiSeq 1000", [100]),
    "HS20": ("HiSeq 2000", [100]),
    "HS25": ("HiSeq 2500", [125, 150]),
    "HSXn": ("HiSeqX PCR free", [150]),
    "HSXt": ("HiSeqX TruSeq", [150]),
    "MinS": ("MiniSeq TruSeq", [50]),
    "MSv1": ("MiSeq v1", [250]),
    "MSv3": ("MSv3 - MiSeq v3", [250]),
    "NS50": ("NextSeq500 v2", [75]),
}
"""
TODO docs

.. versionadded:: 3.1.5
"""

_lh = get_logger(__name__)


class ArtAdapter(BaseProcessBasedLLRGAdapter):
    """
    Wrapper of ART.

    Cmdline Specs::

        cmd = [
            llrg_executable_path,
            "--fcov", str(self._depth),
            "--in", self._src_fasta_file_path,
            "--samout",
            "--noALN",
            "-l", rlen,
            "-O", os.path.join(self._tmp_dir, "tmp"),
            *other_args
        ]

    .. versionadded:: 3.1.5
    """

    def _pre_execution_hook(self) -> None:
        """Does not need extra preparation"""
        pass

    llrg_name: Final[str] = "art"
    _require_integer_depth: Final[bool] = False
    _capture_stdout: Final[bool] = False
    _is_pair_end: bool

    @staticmethod
    def validate_params(
        sequencer_name: str,
        read_length: int,
        pair_end_fragment_length_mean: int,
        pair_end_fragment_length_std: int,
        is_pair_end: bool,
        **kwargs,
    ) -> Mapping[str, Any]:
        _ = kwargs
        retd = {}
        try:
            sequencer_profile = AVAILABLE_ILLUMINA_ART_SEQUENCER[sequencer_name]
        except KeyError as e:
            raise LLRGInitializationException(f"Sequencer Profile for sequencer {sequencer_name} not found!") from e
        if read_length not in sequencer_profile[1]:
            read_length_used = sequencer_profile[1][0]
            _lh.warning(
                "ART: Read length %d not allowed, would use default %d.",
                read_length,
                read_length_used,
            )
        else:
            read_length_used = read_length
        retd.update({"read_length": read_length_used})
        if is_pair_end:
            if pair_end_fragment_length_mean * pair_end_fragment_length_std == 0:
                raise LLRGInitializationException(
                    "Please set pair_end_fragment_length_mean and pair_end_fragment_length_std in PE simulation"
                )
        else:
            if pair_end_fragment_length_mean != 0 or pair_end_fragment_length_std != 0:
                _lh.warning(
                    "ART: PE Params pair_end_fragment_length_mean pair_end_fragment_length_std "
                    "should not be used in SE mode, ignored."
                )
        return retd

    def __init__(
        self,
        *,
        src_fasta_file_path: str,
        dst_fastq_file_prefix: str,
        depth: Union[int, float],
        llrg_executable_path: str,
        is_trusted: bool,
        sequencer_name: str,
        read_length: int,
        pair_end_fragment_length_mean: int,
        pair_end_fragment_length_std: int,
        is_pair_end: bool,
        preserve_intermediate_files: bool,
        other_args: List[str],
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param llrg_executable_path: Path to LLRG Executable.
        :param is_trusted: Whether to skip input validation test.
        :param sequencer_name: name of sequencer to simulate.
        :param read_length: Length of read. Should be defined in corresponding sequencer.
        :param pair_end_fragment_length_mean: [PE Only] The mean size of DNA/RNA fragments for paired-end simulations.
        :param pair_end_fragment_length_std: [PE Only] The standard deviation of DNA/RNA fragment size for
            paired-end simulations.
        :param is_pair_end: Whether the simulation should be pair end.
        :param other_args: Other arguments to be appended at the bottom of assembled CMD.
        :raise LLRGInitializationException: On error.
        """
        super().__init__(
            src_fasta_file_path=src_fasta_file_path,
            dst_fastq_file_prefix=dst_fastq_file_prefix,
            depth=depth,
            llrg_executable_path=llrg_executable_path,
            is_trusted=is_trusted,
            preserve_intermediate_files=preserve_intermediate_files,
        )

        if not is_trusted:
            validated_params = ArtAdapter.validate_params(
                sequencer_name=sequencer_name,
                read_length=read_length,
                pair_end_fragment_length_mean=pair_end_fragment_length_mean,
                pair_end_fragment_length_std=pair_end_fragment_length_std,
                is_pair_end=is_pair_end,
            )
            read_length = validated_params["read_length"]
        self._is_pair_end = is_pair_end
        self._cmd = [
            self._llrg_executable_path,
            "--fcov",
            str(self._depth),
            "--in",
            self._src_fasta_file_path,
            "--samout",
            "--noALN",
            "-l",
            str(read_length),
            "--out",
            os.path.join(self._tmp_dir, "tmp"),
            *other_args,
        ]
        if self._is_pair_end:
            self._cmd.extend(
                [
                    "-p",
                    "--sdev",
                    str(pair_end_fragment_length_std),
                    "--mflen",
                    str(pair_end_fragment_length_mean),
                ]
            )

    def _post_execution_hook(self):
        if self._is_pair_end:
            autocopy(
                os.path.join(self._tmp_dir, "tmp1.fq"),
                self._dst_fastq_file_prefix + "_1.fq",
            )
            autocopy(
                os.path.join(self._tmp_dir, "tmp2.fq"),
                self._dst_fastq_file_prefix + "_2.fq",
            )
        else:
            autocopy(
                os.path.join(self._tmp_dir, "tmp.fq"),
                self._dst_fastq_file_prefix + ".fq",
            )

    @property
    def is_pair_end(self) -> bool:
        return self._is_pair_end


def patch_frontend_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """
    Patch argument parser with ART arguments.

    .. versionadded:: 3.1.5
    """
    parser.add_argument(
        "--sequencer_name",
        required=False,
        help="Name of Illumina Sequencer to Simulate: "
        + ", ".join(
            (
                f"{name} -- {AVAILABLE_ILLUMINA_ART_SEQUENCER[name][0]}"
                for name in AVAILABLE_ILLUMINA_ART_SEQUENCER.keys()
            )
        ),
        nargs="?",
        choices=AVAILABLE_ILLUMINA_ART_SEQUENCER.keys(),
        type=str,
        action="store",
        default="HS25",
    )
    parser.add_argument(
        "--read_length",
        required=False,
        help="Read length. Sequencer -- Read Length Table: "
        + ", ".join((f"{v[0]} -- {v[1]}" for v in AVAILABLE_ILLUMINA_ART_SEQUENCER.values())),
        nargs="?",
        type=int,
        action="store",
        default=0,
    )
    parser.add_argument(
        "--pair_end_fragment_length_mean",
        required=False,
        help="[PE Only] The mean size of DNA/RNA fragments for paired-end simulations",
        nargs="?",
        type=int,
        action="store",
        default=0,
    )
    parser.add_argument(
        "--pair_end_fragment_length_std",
        required=False,
        help="[PE Only] The standard deviation of DNA/RNA fragment size for paired-end simulations.",
        nargs="?",
        type=int,
        action="store",
        default=0,
    )
    parser.add_argument(
        "--is_pair_end",
        required=False,
        help="Whether to use Pair End (PE) Simulation",
        action="store_true",
    )
    parser = patch_frontend_argument_parser(parser, "--preserve_intermediate_files")
    return parser



"""
yasim.llrg_adapter -- Adapters for LLRGs

Here contains LLRG adapters that should have been used in DNA-Seq.

.. versionadded:: 3.1.5
"""
from __future__ import annotations

__all__ = (
    "BaseLLRGAdapter",
    "BaseFunctionBasedLLRGAdapter",
    "BaseProcessBasedLLRGAdapter",
    "EmptyOutputFileException",
    "LLRGInitializationException",
    "LLRGFailException",
    "NoOutputFileException",
    "LLRGException",
    "autocopy",
    "automerge",
    "enhanced_copyfileobj",
)

import logging
import os.path
import subprocess
import threading
from abc import abstractmethod, ABC

from labw_utils.commonutils.lwio import file_system
from labw_utils.commonutils.lwio.safe_io import get_reader, get_writer
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.commonutils.stdlib_helper.shutil_helper import wc_c, rm_rf
from labw_utils.typing_importer import Union, List, IO, Optional, Iterable, Mapping, Any
from yasim.helper.llrg import enhanced_which

COPY_BUFSIZE = 1024 * 1024 if os.name == "nt" else 64 * 1024
"""
Copy buffer size from :py:mod:`shutil`.

.. versionadded:: 3.1.5
"""


class LLRGException(RuntimeError):
    """
    Some error raised due to (improperly configured) LLRG

    .. versionadded:: 3.1.5
    """

    _contents: str

    def __init__(self, contents: str):
        """
        Default Initializer.
        :param contents: What to be put in logs.
        """
        self._contents = contents

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return self._contents


class NoOutputFileException(LLRGException):
    """
    Error raised if no output file was found even when LLRG exited normally.

    .. versionadded:: 3.1.5
    """

    ...


class EmptyOutputFileException(LLRGException):
    """
    Error raised if empty output file was found even when LLRG exited normally.

    .. versionadded:: 3.1.5
    """

    ...


class LLRGFailException(LLRGException):
    """
    Error raised LLRG exited abnormally.

    .. versionadded:: 3.1.5
    """

    ...


class LLRGInitializationException(LLRGException):
    """
    Error raised in initialization or pre-execution stage.

    .. versionadded:: 3.1.5
    """

    ...


def enhanced_copyfileobj(src_fd: IO, dst_fd: IO) -> int:
    """
    :py:func:`shutil.copyfileobj` with number of bytes copied.

    .. versionadded:: 3.1.5
    """
    c_len = 0
    while True:
        buf = src_fd.read(COPY_BUFSIZE)
        if not buf:
            break
        dst_fd.write(buf)
        c_len += len(buf)
    return c_len


def autocopy(in_fn: str, out_fn: str) -> None:
    """
    Copy one file to another place, with automatic extraction of GZipped files

    :param in_fn: Input filename.
    :param out_fn: Output filename.
    :raises LLRGException: If error occurs in copying.

    .. versionadded:: 3.1.5
    """
    try:
        with get_reader(in_fn, is_binary=True) as r1, get_writer(out_fn, is_binary=True) as w1:
            c_len = enhanced_copyfileobj(r1, w1)
    except FileNotFoundError as e:
        raise NoOutputFileException(f"Copy file {in_fn} not found!") from e
    except (OSError, IOError) as e:
        raise LLRGException(f"Copy file {in_fn} -> {out_fn} failed!") from e
    if c_len == 0:
        raise EmptyOutputFileException(f"Copy file {in_fn} empty!")


def automerge(in_fns: Iterable[str], out_fn: str) -> None:
    """
    Merge multiple files into one. See :py:func:`autocopy`.

    .. versionadded:: 3.1.5
    """
    c_len = 0
    with get_writer(out_fn, is_binary=True) as w1:
        for in_fn in in_fns:
            try:
                with get_reader(in_fn, is_binary=True) as r1:
                    c_len += enhanced_copyfileobj(r1, w1)
            except FileNotFoundError as e:
                raise NoOutputFileException(f"Copy file {in_fn} not found!") from e
            except (OSError, IOError) as e:
                raise LLRGException(f"Copy file {in_fn} -> {out_fn} failed!") from e
    if c_len == 0:
        raise EmptyOutputFileException(f"Copy file {in_fns} empty!")


class BaseLLRGAdapter(threading.Thread):
    """
    Base class of LLRG Python adapter.

    The LLRG adapter is a thread-safe Python wrapper of various LLRGs.
    They can be used to simulate DNA and RNA data with pre-defined arguments.


    It performs following operations:

    - Assemble the command that calls LLRG.
    - Call LLRG, see subclasses.
    - Move generated files to ``_dst_fastq_file_prefix``.

    .. versionadded:: 3.1.5
    """

    _src_fasta_file_path: str
    _dst_fastq_file_prefix: str
    _depth: Union[int, float]

    # Following fields are left for LLRGs.
    _lh: logging.Logger
    """
    Class logger handler
    """

    _exception: Optional[LLRGException]
    """Exception raised during initialization, pre-execution, execution or post-execution of LLRG."""

    _preserve_intermediate_files: Optional[bool]
    """
    Whether to preserve intermediate files.
    
    Valid values:
        - True, for preserving them.
        - False, for not preserving them.
        - None, for not creating temporary directory.
    """

    _tmp_dir: Optional[str]
    """
    Simulator-based temp directory name.
    """

    # Following fields are left for LLRG class variables..
    llrg_name: str
    """Class attribute, indicating name. Should be Final."""

    _require_integer_depth: bool
    """Class attribute, indicating whether the depth should be converted to integer. Should be Final."""

    @staticmethod
    def validate_base_params(depth: Union[int, float], src_fasta_file_path: str, **kwargs) -> Mapping[str, Any]:
        _ = kwargs
        if depth <= 0:
            raise LLRGInitializationException(f"Depth {depth} too low")
        if not file_system.file_exists(src_fasta_file_path):
            raise LLRGInitializationException(f"FASTA {src_fasta_file_path} not found!")
        return {}

    @staticmethod
    @abstractmethod
    def validate_params(**kwargs):
        raise NotImplementedError

    def __init__(
        self,
        *,
        src_fasta_file_path: str,
        dst_fastq_file_prefix: str,
        depth: Union[int, float],
        is_trusted: bool,
        preserve_intermediate_files: Optional[bool],
        **kwargs,
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param kwargs: Other miscellaneous arguments.
        :param is_trusted: Whether to skip input validation test.
        :param preserve_intermediate_files: Whether to preserve intermediate files.
            Only effective when LLRG supports temporary file.
        :raise LLRGInitializationException: On error.
        """
        # To developers: This function should raise LLRGInitializationException only!
        super(BaseLLRGAdapter, self).__init__()

        # Check for whether class variables have been properly set.
        if not hasattr(self, "llrg_name"):
            raise TypeError
        if not hasattr(self, "_require_integer_depth"):
            raise TypeError

        if not is_trusted:
            BaseLLRGAdapter.validate_base_params(src_fasta_file_path=src_fasta_file_path, depth=depth)
        self._src_fasta_file_path = src_fasta_file_path
        self._dst_fastq_file_prefix = os.path.abspath(dst_fastq_file_prefix)
        self._depth = int(depth) if self._require_integer_depth else depth
        self._lh = get_logger(__name__)
        self._preserve_intermediate_files = preserve_intermediate_files
        self._exception = None

        if self._preserve_intermediate_files is not None:
            self._tmp_dir = self._dst_fastq_file_prefix + ".tmp.d"
        else:
            self._tmp_dir = None

    @property
    def exception(self) -> str:
        """
        Exception raised during initialization, pre-execution, execution or post-execution of LLRG.

        :return: Following values:

            - ``NORMAL``: If no exception occurs
            - ``EmptyOutFile``: If LLRG exited normally but with empty output file.
            - ``NoOutputFile``: If LLRG exited normally but with no output file.
            - ``LLRGFail``: If LLRG exited abnormally.
            - ``InitFail``: if pre-execution of initialization hook failed.
            - ``UNKNOWN``: Other errors.
        """
        if self._exception is None:
            return "NORMAL"
        elif isinstance(self._exception, EmptyOutputFileException):
            return "EmptyOutFile"
        elif isinstance(self._exception, NoOutputFileException):
            return "NoOutputFile"
        elif isinstance(self._exception, LLRGFailException):
            return "LLRGFail"
        elif isinstance(self._exception, LLRGInitializationException):
            return "InitFail"
        else:
            return "UNKNOWN"

    @abstractmethod
    def _pre_execution_hook(self) -> None:
        """
        Additional steps to take before starting the simulator process

        :raises LLRGException: If error occurs in LLRG-level.
        """
        raise NotImplementedError

    @abstractmethod
    def _post_execution_hook(self):
        """
        Move the file into desired destination.
        May involve (de-)compression.

        :raises LLRGException: If error occurs in LLRG-level.
        """
        raise NotImplementedError

    @abstractmethod
    def _run_llrg_hook(self) -> None:
        """
        Execute LLRG as a process

        :raise LLRGFailException: On running failures.
        """
        raise NotImplementedError

    def run(self) -> None:
        """Execute LLRG"""
        try:
            if self._tmp_dir is not None:
                try:
                    os.makedirs(self._tmp_dir, exist_ok=True)
                except OSError as e:
                    raise LLRGInitializationException("MKTEMP Failed!") from e
            self._pre_execution_hook()
        except LLRGException as e:
            self._lh.warning("LLRG: Exception %s caught at pre-execution time", str(e))
            self._exception = e
            return
        try:
            self._run_llrg_hook()
        except LLRGException as e:
            self._lh.warning("LLRG: Exception %s caught at execution time", str(e))
            self._exception = e
            return
        try:
            self._post_execution_hook()
            if self._preserve_intermediate_files is False:
                try:
                    rm_rf(self._tmp_dir)
                except OSError as e:
                    raise LLRGInitializationException("RMTEMP Failed!") from e
        except LLRGException as e:
            self._lh.warning("LLRG: Exception %s caught at post-execution hook", str(e))
            self._exception = e
            return

    def __repr__(self) -> str:
        return f"LLRG {self.llrg_name}: {self._src_fasta_file_path} -> {self._dst_fastq_file_prefix} [{self._depth}]"

    def __str__(self):
        return repr(self)

    @property
    def input_fasta(self) -> str:
        """
        Input reference FASTA.

        - DNA: Can contain more than one entry.
        - Isoform cDNA: Should contain only one entry.
        """
        return self._src_fasta_file_path

    @property
    def output_fastq_prefix(self) -> str:
        """
        Prefix for output FASTQs.

        For LLRG that generates pair end (PE) reads,
        the generated read will be named ``{_dst_fastq_file_prefix}_1.fq`` and ``{_dst_fastq_file_prefix}_2.fq``.

        For LLRG that generates single-end (SE) reads,
        the generated read will be named ``{_dst_fastq_file_prefix}.fq``
        """
        return self._dst_fastq_file_prefix

    @property
    def depth(self) -> Union[int, float]:
        """
        Sequencing depth or coverage.

        .. warning:: This is NOT final read count!
        """
        return self._depth

    @property
    @abstractmethod
    def is_pair_end(self) -> bool:
        """
        Whether this simulator is pair-end.
        """
        raise NotImplementedError


class BaseProcessBasedLLRGAdapter(BaseLLRGAdapter, ABC):
    """
    LLRG adapters that executes a process using :py:mod:`subprocess`.

    .. versionadded:: 3.1.5
    """

    _llrg_executable_path: str

    # Following fields are left for LLRGs.

    _cmd: Optional[List[str]]
    """Assembled Commandline"""

    _capture_stdout: bool
    """Whether this simulator pours data into stdout. Should be Final."""

    @staticmethod
    def validate_process_based_params(llrg_executable_path: str, **kwargs) -> Mapping[str, Any]:
        _ = kwargs
        try:
            llrg_executable_path = enhanced_which(llrg_executable_path)
        except FileNotFoundError as e:
            raise LLRGInitializationException(
                f"LLRG Executable {llrg_executable_path} not found or not executable!"
            ) from e
        return {"llrg_executable_path": llrg_executable_path}

    def __init__(
        self,
        *,
        src_fasta_file_path: str,
        dst_fastq_file_prefix: str,
        depth: Union[int, float],
        llrg_executable_path: str,
        is_trusted: bool,
        preserve_intermediate_files: Optional[bool],
        **kwargs,
    ):
        """
        Initializer.

        :param src_fasta_file_path: Path of source FASTA.
        :param dst_fastq_file_prefix: Prefix od destination FASTQ.
        :param depth: Targeted sequencing depth. Would NOT be related to actual sequencing depth!
        :param llrg_executable_path: Path to LLRG Executable.
        :param kwargs: Other miscellaneous arguments.
        :param is_trusted: Whether to skip input validation test.
        :param preserve_intermediate_files: Whether to preserve intermediate files.
        :raise LLRGInitializationException: On error.
        """
        # To developers: This function should raise LLRGInitializationException only!
        super(BaseProcessBasedLLRGAdapter, self).__init__(
            src_fasta_file_path=src_fasta_file_path,
            dst_fastq_file_prefix=dst_fastq_file_prefix,
            depth=depth,
            is_trusted=is_trusted,
            preserve_intermediate_files=preserve_intermediate_files,
            **kwargs,
        )

        if not is_trusted:
            validated_params = BaseProcessBasedLLRGAdapter.validate_process_based_params(
                llrg_executable_path=llrg_executable_path
            )
            self._llrg_executable_path = validated_params.get("llrg_executable_path", llrg_executable_path)
        else:
            self._llrg_executable_path = llrg_executable_path
        self._cmd = None

    def _run_llrg_hook(self) -> None:
        """
        Execute LLRG as a process

        :raise LLRGFailException: On running failures.
        """
        if self._cmd is None:
            raise LLRGInitializationException("Commandline Assembly Failed!")
        subprocess_log_file_path = os.path.join(self._tmp_dir, "llrg.log")
        dst_fastq_file_path = self._dst_fastq_file_prefix + ".fq"
        if self._capture_stdout:
            with get_writer(subprocess_log_file_path, is_binary=True) as subprocess_log_handler, get_writer(
                dst_fastq_file_path, is_binary=True
            ) as stdout_handler:
                retv = self._exec_subprocess(
                    self._cmd,
                    stdin=subprocess.DEVNULL,
                    stdout=stdout_handler,
                    stderr=subprocess_log_handler,
                )
            if wc_c(dst_fastq_file_path) < 2:
                raise EmptyOutputFileException(f"Output {dst_fastq_file_path} empty!")
        else:
            with get_writer(subprocess_log_file_path, is_binary=True) as subprocess_log_handler:
                retv = self._exec_subprocess(
                    self._cmd,
                    stdin=subprocess.DEVNULL,
                    stdout=subprocess_log_handler,
                    stderr=subprocess_log_handler,
                )
        if retv != 0:
            raise LLRGFailException(f"Return value LLRG ({retv}) != 0")

    def _exec_subprocess(
        self,
        cmd: List[str],
        stdin: Union[IO, int],
        stdout: Union[IO, int],
        stderr: Union[IO, int],
    ) -> int:
        """
        Wrapper of :py:class:`subprocess.Popen` with logs.
        Will not raise exception if error occurs.

        :param cmd: Commandline arguments.
        :param stdin: File descriptor for STDIN.
        :param stdout: File descriptor for STDOUT.
        :param stderr: File descriptor for STDERR.
        :return: Return value of the process.
        """
        self._lh.debug(f"LLRG: Subprocess {' '.join(cmd)} START")
        retv = subprocess.Popen(cmd, stdin=stdin, stdout=stdout, stderr=stderr).wait()
        if retv == 0:
            self._lh.debug(f"LLRG: Subprocess {' '.join(cmd)} FIN")
        else:
            self._lh.warning(f"LLRG: Subprocess {' '.join(cmd)} ERR={retv}")
        return retv


class BaseFunctionBasedLLRGAdapter(BaseLLRGAdapter, ABC):
    """
    LLRG adapters that executes a function.

    .. versionadded:: 3.1.5
    """

    @abstractmethod
    def llrg_func(self):
        """
        Function that ivokes LLRG and should be overwritten.

        :raises LLRGFailException: on error.
        """
        raise NotImplementedError

    def _run_llrg_hook(self) -> None:
        """
        Execute LLRG as a function

        :raise LLRGFailException: On running failures.
        """
        if self.is_pair_end:
            dst_fastq_file_path1 = self._dst_fastq_file_prefix + ".fq"
            dst_fastq_file_path2 = self._dst_fastq_file_prefix + ".fq"
            self.llrg_func()
            if wc_c(dst_fastq_file_path1) + wc_c(dst_fastq_file_path2) < 2:
                raise EmptyOutputFileException(f"Output {dst_fastq_file_path1} & {dst_fastq_file_path2} empty!")
        else:
            dst_fastq_file_path = self._dst_fastq_file_prefix + ".fq"
            self.llrg_func()
            if wc_c(dst_fastq_file_path) < 2:
                raise EmptyOutputFileException(f"Output {dst_fastq_file_path} empty!")



"""
as_events.py -- Generate AS Events

This file contains necessary modules in *de novo* generation of AS events.
It would reshape Number of Isoforms per Gene on Reference Genome in a log-normal distribution,
with new isoforms introduced by creating AS events and redundant isoforms removed.

.. versionadded:: 3.1.5
"""
from __future__ import annotations

__all__ = ("ASManipulator",)

import array
import random
import uuid

import numpy as np
from scipy.stats import lognorm

from labw_utils.bioutils.datastructure.gene_tree import GeneTreeInterface
from labw_utils.bioutils.datastructure.gv.feature_proxy import BaseFeatureProxy
from labw_utils.bioutils.datastructure.gv.gene import Gene
from labw_utils.bioutils.datastructure.gv.transcript import Transcript
from labw_utils.bioutils.parser.gtf import GtfIteratorWriter
from labw_utils.commonutils.importer.tqdm_importer import tqdm
from labw_utils.commonutils.stdlib_helper.logger_helper import get_logger
from labw_utils.typing_importer import List, Callable, Iterable, Set

ORGANISM_PARAMS = {"ce": (1.0274021145895147, 0.6524307003952217, 0.41902707818534024)}
"""
(s, loc, scale) of a log-normal distribution.

.. versionadded:: 3.1.5
"""

ORGANISM_WEIGHTS = {"ce": (11714, 5886, 8222, 4371)}
"""
Weight of each different AS event type.

.. versionadded:: 3.1.5
"""

_lh = get_logger(__name__)


class SpaceOptimizedTranscript(Transcript):
    __slots__ = ["_exons_bytes"]
    _exons_bytes: bytes

    def __init__(self, t: Transcript):
        self._cdna = None
        self._cdna_unspliced = None
        self._exon_boundaries = None
        self._cdna_unspliced_masked = None
        self._splice_sites = None
        self._is_inferred = t._is_inferred
        self._transcript_id = t.transcript_id  # type: ignore
        self._gene_id = t.gene_id  # type: ignore
        self._exons = list(t.exons)  # Need to copy. # type: ignore
        BaseFeatureProxy.__init__(self, data=t.get_data(), is_checked=t.is_checked)
        self.rescale_from_exon_boundaries(force=True)
        selist = []
        for exon in self.exons:
            selist.append(exon.start0b)
            selist.append(exon.end0b)
        self._exons_bytes = b"".join(
            (
                array.array("Q", selist).tobytes(),
                bytes(self.seqname, encoding="UTF-8"),
                bytes(self.strand),
            )
        )

    def __hash__(self):
        return hash(self._exon_boundaries)

    def __eq__(self, other: "SpaceOptimizedTranscript") -> bool:
        return self._exons_bytes == other._exons_bytes


class ImpossibleToGenerateASEventError(ValueError):
    """
    TODO docs

    .. versionadded:: 3.1.5
    """

    ...


class ASManipulator:
    """
    TODO docs

    .. versionadded:: 3.1.5
    """

    _gv: GeneTreeInterface
    _n_new_isoforms: int
    _n_removed_isoforms: int

    def __init__(self, gv: GeneTreeInterface):
        """
        Creation of an AS manipulator.

        :param gv: GeneTree.

        .. warning::
            The Gene View passed would be internally modified!
        """
        self._gv = gv
        print(self._gv.replace_gene.__qualname__)
        self._n_new_isoforms = 0
        self._n_removed_isoforms = 0

    # @profile
    def _generate_one_as_event(
        self,
        core_func: Callable[[Transcript], Transcript],
        gene: Gene,
        max_try: int,
        sotes: Set[SpaceOptimizedTranscript],
    ) -> Gene:
        for _ in range(max_try):
            transcript = random.choice(list(gene.transcript_values))
            try:
                new_transcript = core_func(transcript)
            except IndexError:
                # lh.debug(f"{core_func.__name__} IndexError")
                continue
            if new_transcript.transcribed_length < 250:
                continue
            sotet = SpaceOptimizedTranscript(new_transcript)
            if sotet in sotes:
                continue
            gene = gene.add_transcript(new_transcript)
            sotes.add(sotet)
            self._n_new_isoforms += 1
            return gene
        _lh.debug("GEN AS EVENT: Gene %s: %d trials exhausted", gene.gene_id, max_try)
        raise ImpossibleToGenerateASEventError

    def _generate_multiple_as_events(
        self,
        core_funcs: Iterable[Callable[[Transcript], Transcript]],
        gene: Gene,
        max_try: int = 100,
    ) -> Gene:
        sotes = set(
            map(
                lambda t: SpaceOptimizedTranscript(t),
                gene.transcript_values,
            )
        )
        for core_func in core_funcs:
            gene = self._generate_one_as_event(
                core_func=core_func,
                gene=gene,
                max_try=max_try,
                sotes=sotes,
            )
        return gene

    @staticmethod
    def _core_perform_exon_skipping(old_transcript: Transcript) -> Transcript:
        number_of_exons = old_transcript.number_of_exons
        # set the percentage of knocked-out exons
        percent = random.random() * 0.3 + 0.3
        # get the number of exons n to be knocked out by multiplying total exon number of transcript
        new_transcript = old_transcript.duplicate(transcript_id=old_transcript.gene_id + str(uuid.uuid4()))
        es_ids = random.sample(range(0, number_of_exons), int(number_of_exons * percent))
        if len(es_ids) == 0:
            raise ImpossibleToGenerateASEventError
        _lh.debug(f"GEN AS EVENT: {new_transcript.transcript_id}: ES {es_ids}, total={number_of_exons}")
        for exon_id in es_ids:
            new_transcript = new_transcript.del_exon(exon_id)
        return new_transcript

    @staticmethod
    def _core_perform_intron_retention(old_transcript: Transcript) -> Transcript:
        number_of_exons = old_transcript.number_of_exons
        start_exon_id = random.choice(range(0, number_of_exons - 2))
        stop_exon_id = start_exon_id + 1
        new_transcript = old_transcript.duplicate(transcript_id=old_transcript.gene_id + str(uuid.uuid4()))
        _lh.debug(
            f"GEN AS EVENT: {new_transcript.transcript_id}: IR {start_exon_id}-{stop_exon_id}, total={number_of_exons}"
        )
        exon_to_change = new_transcript.get_exon(start_exon_id)
        new_transcript = new_transcript.del_exon(start_exon_id)
        exon_to_change = exon_to_change.update_attribute(end=new_transcript.get_exon(stop_exon_id).end)
        new_transcript.add_exon(exon_to_change)
        new_transcript = new_transcript.del_exon(stop_exon_id)
        return new_transcript

    @staticmethod
    def _core_perform_alternative_3p_splicing(old_transcript: Transcript) -> Transcript:
        # randomly pick an exon for splicing
        exon_id = random.randint(0, (old_transcript.number_of_exons - 1))
        # randomly generate the percent to shorten
        exon_len = old_transcript.get_exon(exon_id).transcribed_length
        intron_len = old_transcript.get_intron_length(exon_id)
        splice_perc = (random.random() - 0.5) * 1.6
        delta = int(min(exon_len, intron_len) * splice_perc)
        new_transcript = old_transcript.duplicate(transcript_id=old_transcript.gene_id + str(uuid.uuid4()))
        _lh.debug(
            f"GEN AS EVENT: {new_transcript.transcript_id}: "
            f"A3P {exon_id} (EXON_LEN={exon_len}, INTRON_LEN={intron_len}), {delta}"
        )
        exon_to_change = new_transcript.get_exon(exon_id)
        new_transcript = new_transcript.del_exon(exon_id)
        exon_to_change = exon_to_change.update_attribute(end=exon_to_change.end + delta)
        new_transcript.add_exon(exon_to_change)
        return new_transcript

    @staticmethod
    def _core_perform_alternative_5p_splicing(old_transcript: Transcript) -> Transcript:
        # randomly pick an exon for splicing
        exon_id = random.randint(0, old_transcript.number_of_exons - 1)
        # randomly generate the percent to shorten
        exon_len = old_transcript.get_exon(exon_id).transcribed_length
        intron_len = old_transcript.get_intron_length(exon_id - 1)
        splice_perc = (random.random() - 0.5) * 1.6
        delta = int(min(exon_len, intron_len) * splice_perc)
        new_transcript = old_transcript.duplicate(transcript_id=old_transcript.gene_id + str(uuid.uuid4()))
        _lh.debug(
            f"GEN AS EVENT: {new_transcript.transcript_id}: "
            f"A5P {exon_id} (EXON_LEN={exon_len}, INTRON_LEN={intron_len}), {delta}"
        )
        exon_to_change = new_transcript.get_exon(exon_id)
        new_transcript = new_transcript.del_exon(exon_id)
        exon_to_change = exon_to_change.update_attribute(start=exon_to_change.start + delta)
        new_transcript.add_exon(exon_to_change)
        return new_transcript

    def _perform_alternative_splicing(self, gene: Gene, organism: str) -> Gene:
        core_funcs: Iterable[Callable[[str], SpaceOptimizedTranscript]] = random.choices(
            population=(
                (self._core_perform_alternative_3p_splicing,),
                (self._core_perform_alternative_5p_splicing,),
                (self._core_perform_exon_skipping,),
                (self._core_perform_intron_retention,),
            ),
            weights=ORGANISM_WEIGHTS[organism],
            k=1,
        )[0]
        return self._generate_multiple_as_events(core_funcs, gene)

    def _try_generate_n_isoform_for_a_gene(self, gene: Gene, n: int, organism: str):
        transcript_ids_to_del = []
        for transcript in gene.transcript_values:
            if transcript.transcribed_length < 250:
                transcript_ids_to_del.append(transcript.transcript_id)
        for transcript_id in transcript_ids_to_del:
            gene = gene.del_transcript(transcript_id)
            self._n_removed_isoforms += 1
        if gene.number_of_transcripts == 0:
            raise ImpossibleToGenerateASEventError("Gene cannot transcribed to length >= 250")
        elif gene.number_of_transcripts == n:
            pass
        elif gene.number_of_transcripts > n:
            while gene.number_of_transcripts > n:
                gene = gene.del_transcript(gene.transcript_ids[0])
                self._n_removed_isoforms += 1
        elif gene.number_of_transcripts < n:
            number_of_fail = 0
            while True:
                try:
                    gene = self._perform_alternative_splicing(gene, organism)
                except ImpossibleToGenerateASEventError:
                    number_of_fail += 1
                if number_of_fail > 2 * n:
                    raise ImpossibleToGenerateASEventError("Generation FAILED!")
                elif gene.number_of_transcripts == n:
                    break
        self._gv = self._gv.replace_gene(gene)

    def run(self, organism: str, compl_idx: int) -> None:
        """
        Execute the ASManipulator.

        :param organism: Name of the organism. Allowed are:

            - ``ce``: *C. Elegans*.
        :param compl_idx: The Transcriptome Complexity Index. Larger for larger Number of Isoforms in a Gene.
        """

        _lh.info(
            "GEN AS EVENT: Loaded %d genes with %d transcript",
            self._gv.number_of_genes,
            self._gv.number_of_transcripts,
        )
        if organism not in ORGANISM_PARAMS.keys():
            raise ValueError(f"no organism {organism} available! Available: {ORGANISM_PARAMS.keys()}")
        fit_tuple = ORGANISM_PARAMS[organism]
        gene_ids_to_del: List[str] = []
        all_gene_ids = list(self._gv.gene_ids)
        targeted_nipg = np.array(
            lognorm.rvs(
                fit_tuple[0],
                loc=fit_tuple[1],
                scale=fit_tuple[2],
                size=len(all_gene_ids) * 5,
            ),
            dtype=float,
        )
        minvalue = np.min(targeted_nipg)
        targeted_nipg = (targeted_nipg - minvalue) * compl_idx / np.mean(targeted_nipg) + minvalue
        targeted_nipg = targeted_nipg[np.logical_and(1 < targeted_nipg, targeted_nipg < 25)][: len(all_gene_ids)]
        targeted_nipg = np.array(np.sort(targeted_nipg), dtype=int)
        reference_nipg = [self._gv.get_gene(gene_id)[0].number_of_transcripts for gene_id in all_gene_ids]
        reference_nipg_rank = np.argsort(reference_nipg)
        for gene_index in tqdm(
            iterable=range(len(all_gene_ids)),
            desc="Generating isoforms...",
            total=self._gv.number_of_genes,
        ):
            gene = self._gv.get_gene(all_gene_ids[gene_index])[0]
            n = targeted_nipg[reference_nipg_rank[gene_index]].item()
            try:
                self._try_generate_n_isoform_for_a_gene(gene, n, organism)
            except ImpossibleToGenerateASEventError:
                gene_ids_to_del.append(gene.gene_id)
                self._n_removed_isoforms += gene.number_of_transcripts

        _lh.info("GEN AS EVENT: Removing genes that failed to produce enough isoforms...")
        for gene_id in tqdm(gene_ids_to_del, desc="Removing genes"):
            self._gv = self._gv.del_gene(gene_id)
        _lh.info(
            "GEN AS EVENT: Finished with %d genes (-%d) and %d transcript (+%d, -%d)",
            self._gv.number_of_genes,
            len(gene_ids_to_del),
            self._gv.number_of_transcripts,
            self._n_new_isoforms,
            self._n_removed_isoforms,
        )

    def to_file(self, dst_gtf_file_path: str) -> None:
        """
        Write the internal GeneView to GTF.

        :param dst_gtf_file_path: Path to GTF to be written.
        """
        GtfIteratorWriter.write_iterator(self._gv.to_feature_iterator(), dst_gtf_file_path)



"""
self_check.py -- Check whether YASIM installation is complete.

.. versionadded:: 3.1.6
"""
import importlib
import shutil

from labw_utils.typing_importer import List, Callable, Any


def get_version(
    pkg_name: str,
    version_hook: Callable[[Any], str] = lambda m: getattr(m, "__version__"),
):
    try:
        mod = importlib.import_module(pkg_name)
        ver = version_hook(mod)
    except (
        ImportError,
        ValueError,
        AttributeError,
        RuntimeError,
        SystemError,
        TypeError,
    ):
        ver = "ERR"
    print(f"{pkg_name}: {ver}")


def get_exec(exec_name: str):
    exec_path = shutil.which(exec_name)
    if exec_path is None:
        exec_path = "ERR"
    print(f"{exec_name}: {exec_path}")


def main(args: List[str]):
    _ = args
    del args

    print(f"YASIM version info:")
    for pkg_name in [
        "labw_utils",
        "yasim",
        "pandas",
        "numpy",
        "scipy",
        "tqdm",
        "joblib",
        "pysam",
        "jinja2",
        "h5py",
        "setuptools",
        "seaborn",
    ]:
        get_version(pkg_name)
    get_version("matplotlib", lambda m: m._version.version)

    for exec_name in [
        "badread",
        "pbsim",
        "pbsim2",
        "pbsim3",
        "dwgsim",
        "art_illumina",
        "ccs",
        "samtools",
    ]:
        get_exec(exec_name)



"""
frontend.py -- Utilities for other ``main`` functions.

.. versionadded:: 3.1.5
"""

__all__ = ("patch_frontend_argument_parser",)

import argparse

from labw_utils.bioutils.comm_frontend_opts import FrontendOptSpecs, FrontendOptSpec
from yasim.helper import depth

FrontendOptSpecs.add(
    FrontendOptSpec(
        "-d",
        "--depth",
        required=True,
        help="Path to input Isoform-Level Depth TSV. Can be compressed.",
        nargs="?",
        type=str,
        action="store",
    )
)

FrontendOptSpecs.add(
    FrontendOptSpec(
        "-b",
        "--barcodes",
        required=True,
        help="Path to input barcode TXT. Can be compressed.",
        nargs="?",
        type=str,
        action="store",
    )
)
FrontendOptSpecs.add(
    FrontendOptSpec(
        "--low_cutoff",
        required=False,
        help="Depth lower than this value would be this value.",
        nargs="?",
        type=float,
        action="store",
        default=depth.DEFAULT_LOW_CUTOFF,
    )
)
FrontendOptSpecs.add(
    FrontendOptSpec(
        "--high_cutoff_ratio",
        required=False,
        help="Depth higher than `mu * high_cutoff_ratio` would be `mu * high_cutoff_ratio`",
        nargs="?",
        type=float,
        action="store",
        default=depth.DEFAULT_HIGH_CUTOFF_RATIO,
    )
)
FrontendOptSpecs.add(
    FrontendOptSpec(
        "--preserve_intermediate_files",
        required=False,
        help="Do not remove intermediate files.",
        action="store_true",
    )
)


def patch_frontend_argument_parser(parser: argparse.ArgumentParser, argname: str) -> argparse.ArgumentParser:
    """
    TODO docs

    .. versionadded:: 3.1.5
    """
    return FrontendOptSpecs.patch(parser, argname)



from __future__ import annotations

import enum
import json
import random
from abc import abstractmethod

from labw_utils.bioutils.datastructure.fasta_view import (
    FastaViewType,
    normalize_nt_sequence,
)
from labw_utils.bioutils.datastructure.gene_tree import GeneTreeInterface
from labw_utils.bioutils.datastructure.transposon import TransposonDatabase
from labw_utils.bioutils.parser.fasta import FastaWriter
from labw_utils.bioutils.record.fasta import FastaRecord
from labw_utils.commonutils.importer.tqdm_importer import tqdm
from labw_utils.commonutils.lwio.safe_io import get_writer, get_reader
from labw_utils.typing_importer import Tuple, List, Union, Mapping, Any, Dict, Optional
from yasim.helper import depth_io, depth

DEFAULT_WEIGHT_TRANSCRIPT = 80
DEFAULT_WEIGHT_TE = 20
DEFAULT_WEIGHT_STOP = 100
DEFAULT_MINIMAL_SEQ_LEN = 250
DEFAULT_MINIMAL_TRANSPOSON_LEN = 20
DEFAULT_MINIMAL_TRANSCRIPT_LEN = 120
DEFAULT_MAX_N_GENE = 3
DEFAULT_MAX_N_TRANSPOSON = 3
DEFAULT_MAX_N_FEATURE = 4


class SimpleSerializable:
    seq: str

    @classmethod
    @abstractmethod
    def from_dict(cls, d: Mapping[str, Any]):
        raise NotImplementedError

    @abstractmethod
    def to_dict(self) -> Mapping[str, Any]:
        raise NotImplementedError

    def __eq__(self, others: SimpleSerializable) -> bool:
        if isinstance(others, self.__class__):
            raise TypeError
        return self.seq == others.seq

    def __hash__(self) -> int:
        return hash(self.seq)


class SimpleExon(SimpleSerializable):
    src_gene_id: str
    seq: str

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        return cls(**d)

    def to_dict(self):
        return {
            "type": "SimpleExon",
            "src_gene_id": self.src_gene_id,
            "seq": self.seq,
        }

    def __init__(
        self,
        *,
        src_gene_id: str,
        seq: str,
        **kwargs,
    ) -> None:
        _ = kwargs
        del kwargs
        self.src_gene_id = src_gene_id
        self.seq = seq


class SimpleTE(SimpleSerializable):
    src_te_name: str
    seq: str

    def __init__(
        self,
        src_te_name: str,
        seq: str,
        **kwargs,
    ):
        _ = kwargs
        del kwargs
        self.src_te_name = src_te_name
        self.seq = seq

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        return cls(**d)

    def to_dict(self):
        return {
            "type": "SimpleTE",
            "src_te_name": self.src_te_name,
            "seq": self.seq,
        }


class SimpleTranscript(SimpleSerializable):
    l: List[Union[SimpleExon, SimpleTE]]
    depth: float
    _seq: Optional[str]

    def __init__(self, l: List[Union[SimpleExon, SimpleTE]], depth: float) -> None:
        self.l = l
        self.depth = depth
        self._seq = None

    def to_dict(self):
        return {
            "type": "SimpleTranscript",
            "l": {str(k): v.to_dict() for k, v in enumerate(self.l)},
            "d": self.depth,
        }

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        return cls(
            l=[
                SimpleExon.from_dict(v) if v["type"] == "SimpleExon" else SimpleTE.from_dict(v) for v in d["l"].values()
            ],
            depth=d["d"],
        )

    @property
    def seq(self) -> str:
        if self._seq is None:
            self._seq = "".join(it.seq for it in self.l)
        return self._seq


class TranslationInstructionState(enum.Enum):
    TRANSCRIPT = 0
    TRANSPOSON = 1
    STOP = 2


class TranslationInstructionStateAutomata:
    _weights: Tuple[float, float, float]
    _rdg: random.SystemRandom

    def __init__(self, weights: Tuple[float, float, float]):
        self._rdg = random.SystemRandom()
        self._weights = weights

    def draw(self) -> TranslationInstructionState:
        return self._rdg.choices(
            (
                TranslationInstructionState.TRANSCRIPT,
                TranslationInstructionState.TRANSPOSON,
                TranslationInstructionState.STOP,
            ),
            weights=self._weights,
            k=1,
        )[0]


class TranslationInstruction(SimpleSerializable):
    transcripts: Dict[str, SimpleTranscript]

    def __init__(self, transcripts: Dict[str, SimpleTranscript]) -> None:
        self.transcripts = transcripts

    def to_dict(self):
        return {k: v.to_dict() for k, v in self.transcripts.items()}

    def to_fasta(self, dst_fasta_path: str):
        with FastaWriter(dst_fasta_path) as faw:
            for k, v in self.transcripts.items():
                faw.write(FastaRecord(seq_id=k, sequence=v.seq))

    def to_json(self, dst_json_path: str):
        with get_writer(dst_json_path, is_binary=False) as w:
            json.dump(self.to_dict(), w, indent=4)

    def to_depth(self, dst_depth_path: str):
        depth_data = {k: v.depth for k, v in self.transcripts.items()}
        depth_io.write_depth(depth_data, dst_depth_path, feature_name="TRANSCRIPT_ID")

    @classmethod
    def from_dict(cls, d: Dict[str, Any]):
        return cls({k: SimpleTranscript.from_dict(v) for k, v in d.items()})

    @classmethod
    def from_json(cls, src_json_file_path: str):
        with get_reader(src_json_file_path, is_binary=False) as r:
            return cls.from_dict(json.load(r))

    @property
    def seq(self) -> str:
        raise ValueError

    @classmethod
    def generate(
        cls,
        *,
        n: int,
        tedb: Optional[TransposonDatabase],
        gt: GeneTreeInterface,
        fav: FastaViewType,
        mu: float = depth.DEFAULT_MU,
        disable_gmm: bool = False,
        weight_transcript: float = DEFAULT_WEIGHT_TRANSCRIPT,
        weight_transposon: float = DEFAULT_WEIGHT_TE,
        weight_stop: float = DEFAULT_WEIGHT_STOP,
        minimal_seq_len: int = DEFAULT_MINIMAL_SEQ_LEN,
        minimal_transposon_len: int = DEFAULT_MINIMAL_TRANSPOSON_LEN,
        minimal_transcript_len: int = DEFAULT_MINIMAL_TRANSCRIPT_LEN,
        high_cutoff_ratio: float = depth.DEFAULT_HIGH_CUTOFF_RATIO,
        low_cutoff: float = depth.DEFAULT_LOW_CUTOFF,
        max_n_gene: int = DEFAULT_MAX_N_GENE,
        max_n_transposon: int = DEFAULT_MAX_N_TRANSPOSON,
        max_n_feature: int = DEFAULT_MAX_N_FEATURE,
    ):
        rdg = random.SystemRandom()

        def autoclip(_seq: str, _min_len: int) -> str:
            while True:
                start = rdg.randint(0, len(seq) - 1)
                end = rdg.randint(start, len(seq))
                if end - start + 1 > _min_len:
                    return _seq[start:end]

        collapsed_transcripts = [gene.collapse_transcript(True) for gene in tqdm(gt.gene_values, "Collapsing...")]
        tisa = TranslationInstructionStateAutomata((weight_transcript, weight_transposon, weight_stop))
        final_simple_transcripts: Dict[str, SimpleTranscript] = {}
        pbar = tqdm(desc="Generating sequences...", total=n)
        while len(final_simple_transcripts) < n:
            new_transcript = SimpleTranscript(l=[], depth=0)
            n_gene = 0
            n_transposon = 0
            while n_gene < max_n_gene and n_transposon < max_n_transposon and (n_gene + n_transposon) < max_n_feature:
                state = tisa.draw()
                if state == TranslationInstructionState.STOP:
                    break
                elif state == TranslationInstructionState.TRANSCRIPT:
                    transcript_to_use = rdg.choice(collapsed_transcripts)
                    seq = transcript_to_use.transcribe(fav.sequence, fav.legalize_region_best_effort)
                    if len(seq) < 2 * minimal_transcript_len:
                        continue
                    seq = autoclip(seq, minimal_transcript_len)
                    seq = normalize_nt_sequence(
                        seq,
                        force_upper_case=True,
                        convert_u_into_t=True,
                        convert_non_agct_to_n=True,
                        n_operation="random_assign",
                    )
                    new_transcript.l.append(SimpleExon(src_gene_id=transcript_to_use.gene_id, seq=seq))
                    n_gene += 1
                elif state == TranslationInstructionState.TRANSPOSON and tedb is not None:
                    src_te_name, seq = tedb.draw()
                    if len(seq) < 2 * minimal_transposon_len:
                        continue
                    seq = autoclip(seq, minimal_transposon_len)
                    seq = normalize_nt_sequence(
                        seq,
                        force_upper_case=True,
                        convert_u_into_t=True,
                        convert_non_agct_to_n=True,
                        n_operation="random_assign",
                    )
                    new_transcript.l.append(SimpleTE(src_te_name=src_te_name, seq=seq))
                    n_transposon += 1

            if len(new_transcript.seq) < minimal_seq_len:
                continue
            final_simple_transcripts[f"gtt-{len(final_simple_transcripts)}"] = new_transcript
            pbar.update(1)
        if disable_gmm:
            for v in final_simple_transcripts.values():
                v.depth = mu
        else:
            for k, v in depth.simulate_gene_level_depth_gmm(
                gene_names=final_simple_transcripts.keys(),
                mu=mu,
                low_cutoff=low_cutoff,
                high_cutoff_ratio=high_cutoff_ratio,
            ).items():
                final_simple_transcripts[k].depth = v
        return cls(final_simple_transcripts)



"""
TODO docs

.. versionadded:: 3.1.5
"""

from labw_utils.commonutils.libfrontend import setup_frontend

from yasim import __version__
from yasim_scripts import description

if __name__ == "__main__":
    setup_frontend(f"{__package__}._main", description, __version__)



"""
generate_isoform_replicates.py -- Generate Technical Replicates using YASIM V3 API.

.. versionadded:: 3.1.5
"""

__all__ = ("create_parser", "main")

import argparse

from labw_utils.commonutils.stdlib_helper.argparse_helper import (
    ArgumentParserWithEnhancedFormatHelp,
)
from labw_utils.typing_importer import List
from yasim.helper import depth, depth_io
from yasim.helper.frontend import patch_frontend_argument_parser


def create_parser() -> argparse.ArgumentParser:
    parser = ArgumentParserWithEnhancedFormatHelp(
        prog="python -m yasim generate_isoform_replicates",
        description=__doc__.splitlines()[1],
    )
    parser = patch_frontend_argument_parser(parser, "-d")
    parser.add_argument(
        "-n",
        "--num_replicates",
        required=False,
        help="Number of Replicates to be generated",
        nargs="?",
        type=int,
        action="store",
        default=3,
    )
    parser.add_argument(
        "-r",
        "--range",
        required=False,
        help="Range of Generated Data",
        nargs="?",
        type=float,
        action="store",
        default=0.1,
    )
    return parser


def main(args: List[str]):
    args = create_parser().parse_args(args)
    depth_data = depth_io.read_depth(args.depth)
    for i in range(args.num_replicates):
        depth_io.write_depth(
            depth.generate_depth_replicates_uniform(depth_data, args.range),
            dst_depth_file_path=f"{args.depth}.{i}",
            feature_name="TRANSCRIPT_ID",
        )
